#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import sqlite3
import progressbar
import pyjson5 as json5
import os.path

from prjxray.util import OpenSafeFile

def create_tables(conn):
    c = conn.cursor()

    c.execute(
        """CREATE TABLE tile(
    pkey INTEGER PRIMARY KEY,
    name TEXT
    );""")
    c.execute(
        """CREATE TABLE node(
    pkey INTEGER PRIMARY KEY,
    name TEXT
    );""")
    c.execute(
        """CREATE TABLE wire(
    pkey INTEGER PRIMARY KEY,
    name TEXT,
    node_pkey INTEGER,
    tile_pkey INTEGER,
    FOREIGN KEY(node_pkey) REFERENCES node(pkey),
    FOREIGN KEY(tile_pkey) REFERENCES tile(pkey)
    );""")

    conn.commit()


class NodeLookup(object):
    def __init__(self, database):
        self.conn = sqlite3.connect(database)

    def build_database(self, nodes, tiles):
        create_tables(self.conn)

        c = self.conn.cursor()
        tile_names = []
        for tile_type in tiles:
            for tile in tiles[tile_type]:
                tile_names.append(tile)

        tile_pkeys = {}
        for tile_file in progressbar.progressbar(tile_names):
            # build/specimen_001/tile_DSP_L_X34Y145.json5
            root, _ = os.path.splitext(os.path.basename(tile_file))
            tile = root[5:]
            c.execute("INSERT INTO tile(name) VALUES (?);", (tile, ))
            tile_pkeys[tile] = c.lastrowid

        nodes_processed = set()
        for node in progressbar.progressbar(nodes):
            with OpenSafeFile(node) as f:
                node_wires = json5.load(f)
                assert node_wires['node'] not in nodes_processed
                nodes_processed.add(node_wires['node'])

                c.execute(
                    "INSERT INTO node(name) VALUES (?);",
                    (node_wires['node'], ))
                node_pkey = c.lastrowid

                for wire in node_wires['wires']:
                    tile = wire['wire'].split('/')[0]

                    tile_pkey = tile_pkeys[tile]
                    c.execute(
                        """
INSERT INTO wire(name, tile_pkey, node_pkey) VALUES (?, ?, ?);""",
                        (wire['wire'], tile_pkey, node_pkey))

        self.conn.commit()

        c = self.conn.cursor()
        c.execute("CREATE INDEX tile_names ON tile(name);")
        c.execute("CREATE INDEX node_names ON node(name);")
        c.execute("CREATE INDEX wire_node_tile ON wire(node_pkey, tile_pkey);")
        c.execute("CREATE INDEX wire_tile ON wire(tile_pkey);")
        self.conn.commit()

    def site_pin_node_to_wires(self, tile, node):
        if node is None:
            return

        c = self.conn.cursor()
        c.execute(
            """
WITH
    the_tile(tile_pkey) AS (SELECT pkey AS tile_pkey FROM tile WHERE name = ?),
    the_node(node_pkey) AS (SELECT pkey AS node_pkey FROM node WHERE name = ?)
SELECT wire.name FROM wire
    INNER JOIN the_tile ON the_tile.tile_pkey = wire.tile_pkey
    INNER JOIN the_node ON the_node.node_pkey = wire.node_pkey;
""", (tile, node))

        for row in c:
            yield row[0][len(tile) + 1:]

    def wires_for_tile(self, tile):
        c = self.conn.cursor()
        c.execute(
            """
WITH
    the_tile(tile_pkey) AS (SELECT pkey AS tile_pkey FROM tile WHERE name = ?)
SELECT wire.name FROM wire
    INNER JOIN the_tile ON the_tile.tile_pkey = wire.tile_pkey;
""", (tile, ))
        for row in c:
            yield row[0][len(tile) + 1:]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
Generate a primitive to place at every I/O
Unlike CLB tests, the LFSR for this is inside the ROI, not driving it
'''

import os
import random
import sys
#random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog


def gen_iobs():
    '''
    IOB33S: main IOB of a diff pair
    IOB33M: secondary IOB of a diff pair
    IOB33: not a diff pair. Relatively rare (at least in ROI...2 of them?)
    Focus on IOB33S to start
    '''
    for _tile_name, site_name, site_type in util.get_roi().gen_sites(
            #['IOB33', 'IOB33S', 'IOB33M']):
        ['IOB33S']):
        yield site_name, site_type


def write_pins(ports):
    pinstr = ''
    for site, (name, dir_, cell) in sorted(ports.items(), key=lambda x: x[1]):
        # pinstr += 'set_property -dict "PACKAGE_PIN %s IOSTANDARD LVCMOS33" [get_ports %s]' % (packpin, port)
        pinstr += '%s,%s,%s,%s\n' % (site, name, dir_, cell)
    open('params.csv', 'w').write(pinstr)


def run():
    # All possible values
    iosites = {}
    for site_name, site_type in gen_iobs():
        iosites[site_name] = site_type

    # Assigned in this design
    ports = {}
    DIN_N = 0
    DOUT_N = 0

    def remain_sites():
        return set(iosites.keys()) - set(ports.keys())

    def rand_site():
        '''Get a random, unused site'''
        return random.choice(list(remain_sites()))

    def get_site():
        return next(iter(remain_sites()))

    def assign_i(site, name):
        nonlocal DIN_N

        assert site not in ports
        cell = "di_bufs[%u].ibuf" % DIN_N
        DIN_N += 1
        ports[site] = (name, 'input', cell)

    def assign_o(site, name):
        nonlocal DOUT_N

        assert site not in ports
        cell = "do_bufs[%u].obuf" % DOUT_N
        DOUT_N += 1
        ports[site] = (name, 'output', cell)

    # Assign at least one di and one do
    assign_i(get_site(), 'di[0]')
    assign_o(get_site(), 'do[0]')
    # Now assign the rest randomly
    #while len(remain_sites()):
    #        assign_o(rand_site(), 'do[%u]' % DOUT_N)

    #write_pins(ports)

    print(
        '''
`define N_DI %u
`define N_DO %u

module top(input wire [`N_DI-1:0] di, output wire [`N_DO-1:0] do);
    genvar i;

    //Instantiate BUFs so we can LOC them

    wire [`N_DI-1:0] di_buf;
    generate
        for (i = 0; i < `N_DI; i = i+1) begin:di_bufs
            IBUF #(
            ) ibuf(.I(di[i]), .O(di_buf[i]));
        end
    endgenerate

    wire [`N_DO-1:0] do_unbuf;
    generate
        for (i = 0; i < `N_DO; i = i+1) begin:do_bufs
            OBUF #(
            ) obuf(.I(do_unbuf[i]), .O(do[i]));
        end
    endgenerate

    roi roi(.di(di_buf), .do(do_unbuf));
endmodule

//Arbitrary terminate into LUTs
module roi(input wire [`N_DI-1:0] di, output wire [`N_DO-1:0] do);
    genvar i;

    generate
        for (i = 0; i < `N_DI; i = i+1) begin:dis
            (* KEEP, DONT_TOUCH *)
            LUT6 #(
                    .INIT(64'h8000_0000_0000_0001)
            ) lut (
                    .I0(di[i]),
                    .I1(di[i]),
                    .I2(di[i]),
                    .I3(di[i]),
                    .I4(di[i]),
                    .I5(di[i]),
                    .O());
        end
    endgenerate

    generate
        for (i = 0; i < `N_DO; i = i+1) begin:dos
            (* KEEP, DONT_TOUCH *)
            LUT6 #(
                    .INIT(64'h8000_0000_0000_0001)
            ) lut (
                    .I0(),
                    .I1(),
                    .I2(),
                    .I3(),
                    .I4(),
                    .I5(),
                    .O(do[i]));
        end
    endgenerate
endmodule
    ''' % (DIN_N, DOUT_N))


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker

segmk = Segmaker("design.bits")

print("Loading tags")
'''
module,loc,bel,n
clb_NCY0_MX,SLICE_X12Y100,A6LUT,3
clb_NCY0_O5,SLICE_X16Y100,C6LUT,0
clb_NCY0_O5,SLICE_X17Y100,A6LUT,2
'''
f = open('params.csv', 'r')
f.readline()
for l in f:
    module, loc, bel, n = l.split(',')
    n = int(n)
    # A, B, etc
    which = bel[0]

    # One bit, set on O5
    segmk.add_site_tag(loc, "CARRY4.%cCY0" % which, module == 'clb_NCY0_O5')
segmk.compile()
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import argparse
import datetime
import progressbar
import os.path
import prjxray.lib
import pickle
import collections
import json


def build_node_index(fname):
    node_index = {}
    with open(fname, 'rb') as f:
        f.seek(0, 2)
        bytes = f.tell()
        f.seek(0, 0)
        with progressbar.ProgressBar(max_value=bytes) as bar:
            end_of_line = 0
            for l in f:
                parts = l.decode('utf8').split(' ')
                pip, node = parts[0:2]

                if node not in node_index:
                    node_index[node] = []

                node_index[node].append(end_of_line)
                end_of_line = f.tell()
                bar.update(end_of_line)

    return node_index


def read_node(expected_node, wire_file, node_index):
    with open(wire_file, 'rb') as f:
        for index in node_index:
            f.seek(index, 0)

            parts = f.readline().decode('utf8').strip().split(' ')

            pip, node = parts[0:2]
            wires = parts[2:]

            assert node == expected_node, repr((node, expected_node, index))

            yield wires


def generate_edges(graph, root, graph_nodes):
    """ Starting from root, generate an edge in dir and insert into graph.

  If the tree forks, simply insert a joins to indicate the split.

  """
    edge = [root]
    prev_root = None

    while True:
        outbound_edges = graph_nodes[root]
        outbound_edges -= set((prev_root, ))
        if len(outbound_edges) > 1:
            graph['edges'].append(edge)
            if root not in graph['joins']:
                graph['joins'][root] = set()
            graph['joins'][root] |= outbound_edges

            for element in graph_nodes[root]:
                if element not in graph['joins']:
                    graph['joins'][element] = set()
                graph['joins'][element].add(root)

            break
        else:
            if len(outbound_edges) == 0:
                graph['edges'].append(edge)
                break

            next_root = tuple(outbound_edges)[0]
            edge.append(next_root)
            prev_root, root = root, next_root


def create_ordered_wires_for_node(node, wires_in_node, downhill, uphill):
    if len(wires_in_node) <= 2:
        return {'edges': [wires_in_node], 'joins': {}}

    downhill = set(tuple(l) for l in downhill)
    uphill = set(tuple(l) for l in uphill)

    roots = set()
    all_wires = set()

    for wire in downhill:
        if len(wire) > 0:
            roots |= set((wire[0], wire[-1]))
            all_wires |= set(wire)

    for wire in uphill:
        if len(wire) > 0:
            roots |= set((wire[0], wire[-1]))
            all_wires |= set(wire)

    assert len(wires_in_node) >= len(all_wires), (
        len(wires_in_node), len(all_wires))

    if len(all_wires) <= 2:
        return {'edges': tuple(all_wires), 'joins': {}}

    graph_nodes = dict((wire, set()) for wire in all_wires)

    for wire in all_wires:
        for down in downhill:
            try:
                idx = down.index(wire)
                if idx + 1 < len(down):
                    graph_nodes[wire].add(down[idx + 1])
                if idx - 1 >= 0:
                    graph_nodes[wire].add(down[idx - 1])
            except ValueError:
                continue

        for up in uphill:
            try:
                idx = up.index(wire)
                if idx + 1 < len(up):
                    graph_nodes[wire].add(up[idx + 1])
                if idx - 1 >= 0:
                    graph_nodes[wire].add(up[idx - 1])
            except ValueError:
                continue

    graph = {'edges': [], 'joins': {}}

    while len(roots) > 0:
        root = roots.pop()

        if len(graph_nodes[root]) > 0:
            generate_edges(graph, root, graph_nodes)

    # Dedup identical edges.
    final_edges = set()

    for edge in graph['edges']:
        edge1 = tuple(edge)
        edge2 = tuple(edge[::-1])

        if edge1 > edge2:
            final_edges.add((edge2, edge1))
        else:
            final_edges.add((edge1, edge2))

    edges = [edge[0] for edge in final_edges]

    element_index = {}
    for edge in edges:
        for idx, element in enumerate(edge):
            if element not in element_index:
                element_index[element] = []
            element_index[element].append((idx, edge))

    new_edges = []
    for edge in edges:
        starts = element_index[edge[0]]
        ends = element_index[edge[-1]]

        found_any = False
        for start in starts:
            start_idx, other_edge = start
            if other_edge is edge:
                continue

            for end in ends:
                if other_edge is not end[1]:
                    continue

                found_any = True
                end_idx, _ = end
                # check if the interior elements are the same.
                if start_idx > end_idx:
                    step = -1
                else:
                    step = 1

                other_edge_slice = slice(
                    start_idx, end_idx + step if end_idx + step >= 0 else None,
                    step)
                if edge != other_edge[other_edge_slice]:
                    new_edges.append(edge)

        if not found_any:
            new_edges.append(edge)

    output = {
        'edges':
        new_edges,
        'joins':
        dict((key, tuple(value)) for key, value in graph['joins'].items()),
        'wires':
        wires_in_node,
    }

    all_wires_in_output = set()
    for edge in output['edges']:
        all_wires_in_output |= set(edge)

    for element in output['joins']:
        all_wires_in_output.add(element)

    return output


def main():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument('--dump_all_root_dir', required=True)
    parser.add_argument('--ordered_wires_root_dir', required=True)
    parser.add_argument('--output_dir', required=True)

    args = parser.parse_args()

    downhill_wires = os.path.join(
        args.ordered_wires_root_dir, 'downhill_wires.txt')
    uphill_wires = os.path.join(
        args.ordered_wires_root_dir, 'uphill_wires.txt')

    assert os.path.exists(downhill_wires)
    assert os.path.exists(uphill_wires)

    print('{} Reading root.csv'.format(datetime.datetime.now()))
    tiles, nodes = prjxray.lib.read_root_csv(args.dump_all_root_dir)

    print('{} Loading node<->wire mapping'.format(datetime.datetime.now()))
    node_lookup = prjxray.lib.NodeLookup()
    node_lookup_file = os.path.join(args.output_dir, 'nodes.pickle')
    if os.path.exists(node_lookup_file):
        node_lookup.load_from_file(node_lookup_file)
    else:
        node_lookup.load_from_root_csv(nodes)
        node_lookup.save_to_file(node_lookup_file)

    wire_index_file = os.path.join(args.output_dir, 'wire_index.pickle')
    if os.path.exists(wire_index_file):
        print('{} Reading wire<->node index'.format(datetime.datetime.now()))
        with open(wire_index_file, 'rb') as f:
            wire_index = pickle.load(f)

        downhill_wire_node_index = wire_index['downhill']
        uphill_wire_node_index = wire_index['uphill']
    else:
        print('{} Creating wire<->node index'.format(datetime.datetime.now()))
        downhill_wire_node_index = build_node_index(downhill_wires)
        uphill_wire_node_index = build_node_index(uphill_wires)

        with open(wire_index_file, 'wb') as f:
            pickle.dump(
                {
                    'downhill': downhill_wire_node_index,
                    'uphill': uphill_wire_node_index,
                }, f)

    print('{} Creating node tree'.format(datetime.datetime.now()))
    nodes = collections.OrderedDict()
    for node in progressbar.progressbar(sorted(node_lookup.nodes)):
        nodes[node] = create_ordered_wires_for_node(
            node, tuple(wire['wire'] for wire in node_lookup.nodes[node]),
            tuple(
                read_node(
                    node, downhill_wires, downhill_wire_node_index[node]
                    if node in downhill_wire_node_index else [])),
            tuple(
                read_node(
                    node, uphill_wires, uphill_wire_node_index[node]
                    if node in uphill_wire_node_index else [])))

    print('{} Writing node tree'.format(datetime.datetime.now()))
    with open(os.path.join(args.output_dir, 'node_tree.json'), 'w') as f:
        json.dump(nodes, f, indent=2, sort_keys=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
#random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        if gridinfo.tile_type in ['CFG_CENTER_MID']:
            for site_name in sorted(gridinfo.sites.keys()):
                if site_name.startswith("BSCAN_X0Y0"):
                    yield tile_name, site_name


def write_params(params):
    pinstr = 'tile,val,site\n'
    for tile, (site, val) in sorted(params.items()):
        pinstr += '%s,%s,%s\n' % (tile, val, site)
    open('params.csv', 'w').write(pinstr)


def run():
    print('''
module top();
    ''')
    params = {}

    sites = list(gen_sites())
    jtag_chains = ("1", "2", "3", "4")
    for (tile_name, site_name), isone in zip(sites,
                                             util.gen_fuzz_states(len(sites))):
        params[tile_name] = (site_name, isone)
        print(
            '''
            (* KEEP, DONT_TOUCH, LOC = "{0}" *)
            BSCANE2 #(
            .JTAG_CHAIN("{1}")
            )
            BSCANE2_{0} (
            .CAPTURE(),
            .DRCK(),
            .RESET(),
            .RUNTEST(),
            .SEL(),
            .SHIFT(),
            .TCK(),
            .TDI(),
            .TMS(),
            .UPDATE(),
            .TDO(1'b1)
            );
        '''.format(site_name, jtag_chains[isone]))

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
import itertools
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


def gen_brams():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        sites = []
        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['RAMB18E1', 'FIFO18E1']:
                sites.append(site_name)

        sites.sort()

        if len(sites) == 0:
            continue

        if gridinfo.tile_type[-1] == 'L':
            int_grid_x = loc.grid_x + 2
            int_tile_type = 'INT_L'
        else:
            int_grid_x = loc.grid_x - 2
            int_tile_type = 'INT_R'

        int_tile_locs = [
            (int_grid_x, loc.grid_y),
            (int_grid_x, loc.grid_y - 1),
            (int_grid_x, loc.grid_y - 2),
            (int_grid_x, loc.grid_y - 3),
            (int_grid_x, loc.grid_y - 4),
        ]

        int_tiles = []
        for int_tile_loc in int_tile_locs:
            int_gridinfo = grid.gridinfo_at_loc(int_tile_loc)
            assert int_gridinfo.tile_type == int_tile_type, (
                int_gridinfo.tile_type, int_tile_type)

            int_tiles.append(grid.tilename_at_loc(int_tile_loc))

        yield tile_name, sites, int_tiles


def write_params(params):
    pinstr = 'tile,val\n'
    for tile, (val, ) in sorted(params.items()):
        pinstr += '%s,%s\n' % (tile, val)
    open('params.csv', 'w').write(pinstr)


def run():
    print('''
module top();
    ''')

    params = {}

    sites = list(gen_brams())
    fuzz_iter = iter(util.gen_fuzz_states(len(sites) * 5))
    for tile_name, bram_sites, int_tiles in sites:
        # Each BRAM tile has 5 INT tiles.
        # The following feature is used to toggle a one PIP in each INT tile
        #
        # For BRAM_L_X6Y0, there are the following INT tiles and the feature
        # to toggle
        #
        #  - INT_L_X6Y0, tie bram_sites[0].DIADI[2] = 0, toggle bram_sites[0].DIADI[3]
        #  - INT_L_X6Y1, tie bram_sites[1].ADDRBWRADDR[7] = 0, toggle bram_sites[1].ADDRBWRADDR[10]
        #  - INT_L_X6Y2, tie bram_sites[1].ADDRARDADDR[9] = 0, toggle bram_sites[1].ADDRBWRADDR[9]
        #  - INT_L_X6Y3, tie bram_sites[1].ADDRBWRADDR[4] = 0, toggle bram_sites[1].ADDRBWRADDR[13]
        #  - INT_L_X6Y4, tie bram_sites[1].DIBDI[15] = 0, toggle bram_sites[1].DIADI[7]
        (b0_diadi3, b1_wraddr10, b1_wraddr9, b1_wraddr13,
         b1_diadi7) = itertools.islice(fuzz_iter, 5)
        params[int_tiles[0]] = (b0_diadi3, )
        params[int_tiles[1]] = (b1_wraddr10, )
        params[int_tiles[2]] = (b1_wraddr9, )
        params[int_tiles[3]] = (b1_wraddr13, )
        params[int_tiles[4]] = (b1_diadi7, )

        print(
            '''
            wire [15:0] {bram_site0}_diadi;
            wire [15:0] {bram_site0}_dibdi;
            wire [13:0] {bram_site0}_wraddr;

            wire [15:0] {bram_site1}_diadi;
            wire [15:0] {bram_site1}_dibdi;
            wire [7:0] {bram_site1}_webwe;
            wire [13:0] {bram_site1}_rdaddr;
            wire [13:0] {bram_site1}_wraddr;

            // INT 0
            assign {bram_site0}_diadi[2] = 0;
            assign {bram_site0}_diadi[3] = {b0_diadi3};

            // INT 1
            assign {bram_site1}_wraddr[7] = 0;
            assign {bram_site1}_wraddr[10] = {b1_wraddr10};

            // INT 2
            assign {bram_site1}_rdaddr[9] = 0;
            assign {bram_site1}_wraddr[9] = {b1_wraddr9};

            // INT 3
            assign {bram_site1}_wraddr[4] = 0;
            assign {bram_site1}_wraddr[13] = {b1_wraddr13};

            // INT 4
            assign {bram_site1}_dibdi[15] = 0;
            assign {bram_site1}_diadi[7] = {b1_diadi7};

            (* KEEP, DONT_TOUCH, LOC = "{bram_site0}" *)
            RAMB18E1 #(
                ) bram_{bram_site0} (
                    .CLKARDCLK(),
                    .CLKBWRCLK(),
                    .ENARDEN(),
                    .ENBWREN(),
                    .REGCEAREGCE(),
                    .REGCEB(),
                    .RSTRAMARSTRAM(),
                    .RSTRAMB(),
                    .RSTREGARSTREG(),
                    .RSTREGB(),
                    .ADDRARDADDR(),
                    .ADDRBWRADDR({bram_site0}_wraddr),
                    .DIADI({bram_site0}_diadi),
                    .DIBDI({bram_site0}_dibdi),
                    .DIPADIP(),
                    .DIPBDIP(),
                    .WEA(),
                    .WEBWE(),
                    .DOADO(),
                    .DOBDO(),
                    .DOPADOP(),
                    .DOPBDOP());

            (* KEEP, DONT_TOUCH, LOC = "{bram_site1}" *)
            RAMB18E1 #(
                ) bram_{bram_site1} (
                    .CLKARDCLK(),
                    .CLKBWRCLK(),
                    .ENARDEN(),
                    .ENBWREN(),
                    .REGCEAREGCE(),
                    .REGCEB(),
                    .RSTRAMARSTRAM(),
                    .RSTRAMB(),
                    .RSTREGARSTREG(),
                    .RSTREGB(),
                    .ADDRARDADDR({bram_site1}_rdaddr),
                    .ADDRBWRADDR({bram_site1}_wraddr),
                    .DIADI({bram_site1}_diadi),
                    .DIBDI({bram_site1}_dibdi),
                    .DIPADIP(),
                    .DIPBDIP(),
                    .WEA(),
                    .WEBWE({bram_site1}_webwe),
                    .DOADO(),
                    .DOBDO(),
                    .DOPADOP(),
                    .DOPBDOP());
'''.format(
                bram_site0=bram_sites[0],
                bram_site1=bram_sites[1],
                b0_diadi3=b0_diadi3,
                b1_wraddr10=b1_wraddr10,
                b1_wraddr9=b1_wraddr9,
                b1_wraddr13=b1_wraddr13,
                b1_diadi7=b1_diadi7))

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Description of a site type """
from collections import namedtuple
import enum


class SitePinDirection(enum.Enum):
    IN = "IN"
    OUT = "OUT"
    INOUT = "INOUT"


SiteTypePin = namedtuple('SiteTypePin', 'name direction')


class SiteType(object):
    def __init__(self, site_type):
        self.type = site_type['type']
        self.site_pins = {}
        for site_pin, site_pin_info in site_type['site_pins'].items():
            self.site_pins[site_pin] = SiteTypePin(
                name=site_pin,
                direction=SitePinDirection(site_pin_info['direction']),
            )

    def get_site_pins(self):
        return self.site_pins.keys()

    def get_site_pin(self, site_pin):
        return self.site_pins[site_pin]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
This script solves the fuzzing problem through least-mean-square solution of
an overdetermined linear equation system.

The advantages of this method are:
- Ability to detect negative correlations (tags which require clearing bits)
- Can detect partial correlation tag <-> bit. This happens if for a small
  number of specimens a tag is said to be "1" but in fact it is not due to
  the way Vivado interprets requested features and encodes them into bitstream.
- Ease to detect tags with no corresponding bits by evaluating solution error.

The solution is computed using the Tikhonov regularization scheme to ensure
numerical stability. The parameter -a can be used to vary the regularization
factor.

By default each tag is solved separately (best results) while they can be
solved all at once (not recommended).

For each tag a vector of weights is calculated. Each weight corresponds to one
bit. Positive values indicate positive correlation and negative values negative
correlation.

Each weight vector is normalized so that maximum absolute weight is equal to
one.

The parameter -t is used to set threshold for those weights. Weights with
values above the threshold and below the "minus" threshold are output as
candidate bits.

For each weight vector a solution error is computed. If the error exceeds
threshold specified using the -e parameter then the tag is considered to
have no bits.

The option -m can be used to filter bits found for the specified tag in all
other tags. This allows to remove bits from a "IS_BLOCK_IN_USE" type tag from
other tags responsible for enabling other features of that block.
'''
import sys
import os
import argparse
import itertools
import json

import numpy as np
import numpy.linalg as linalg

from prjxray.util import OpenSafeFile

# =============================================================================


def load_data(file_name, tagfilter=lambda tag: True, address_map=None):
    """
    Loads data generated by the segmaker.

    Parameters
    ----------

    file_name:
        Name of the text file with data.
    tagfilter:
        A function for filtering tags. Should reqturn True or False.
    address_map:
        A dict indexed by tuples (address, offset) containing a list
        of tile names.

    Returns
    -------
        A list of dicts. Each contains:
        - "seg": Segment name
        - "bit": A list of bit names
        - "tag": A list of tuples (tag name, tag value)
    """

    segdata = None
    all_segdata = []

    with OpenSafeFile(file_name, "r") as fp:
        for line in fp.readlines():
            line = line.strip()

            # Segment tag
            if line.startswith("seg"):
                fields = line.split()

                if segdata is not None:
                    if len(segdata["tag"]):
                        all_segdata.append(segdata)
                    segdata = None

                segname = fields[1]

                # Map segment address to tile name
                if address_map is not None:
                    address = segname.split("_")
                    address = (
                        int(address[0], base=16),
                        int(address[1]),
                    )
                    if address in address_map:
                        segname = "_or_".join(address_map[address])

                # Append file name
                segname = file_name + ":" + segname

                # Append segdata
                segdata = {"seg": segname, "bit": [], "tag": []}

            if segdata is None:
                continue

            # Bit tag
            if line.startswith("bit"):
                fields = line.split()
                segdata["bit"].append(fields[1])

            # Tag tag
            if line.startswith("tag"):
                fields = line.split()

                if not tagfilter(fields[1]):
                    continue

                segdata["tag"].append((
                    fields[1],
                    int(fields[2]),
                ))

    # Store the last segment if any
    if segdata is not None:
        if len(segdata["tag"]):
            all_segdata.append(segdata)

    return all_segdata


def write_segbits(file_name, all_tags, all_bits, W):
    """
    Writes solution to a raw database file.

    Parameters
    ----------

    file_name:
        Name of the .rdb file.
    all_tags:
        List of considered tags.
    all_bits:
        List of considered bits.
    W:
        Matrix with binary solution.
    """
    lines = []

    for r in range(W.shape[0]):
        bits = []
        for c in range(W.shape[1]):
            w = W[r, c]
            if w < 0:
                bits.append("!" + all_bits[c])
            if w > 0:
                bits.append(all_bits[c])

        if len(bits) == 0:
            bits = ["<0 candidates>"]

        lines.append(all_tags[r] + " " + " ".join(bits) + "\n")

    with OpenSafeFile(file_name, "w") as fp:
        for line in lines:
            fp.write(line)


def dump_results(fp, all_tags, all_bits, W, X, E, tag_stats=None):
    """
    Dumps solution results to an open file in a nice readable format.

    Parameters
    ----------

    fp:
        An open file or stream
    all_tags:
        List of considered tags.
    all_bits:
        List of considered bits.
    W:
        Matrix with binary solution.
    X:
        Matrix with raw solution (floats).
    E:
        Vector with solution errors.
    tag_stats:
        Tag statistics.
    """
    lines = []

    pad_len = max([len(tag) for tag in all_tags])

    skip_bit = []
    for i in range(len(all_bits)):
        skip_bit.append((W[:, i] == 0).all())

    # Bit names
    bit_len = 6
    for i in range(bit_len):
        line = " " * (pad_len + 2 + 3)
        for j in range(len(all_bits)):
            if skip_bit[j]:
                continue
            bname = all_bits[j].ljust(bit_len).replace("_", "|")
            line += bname[i]

        if i == (bit_len - 1):
            if tag_stats is not None:
                line += "  #0   #1 "

        lines.append(line)

    # Tags and bit values
    pad = max([len(tag) for tag in all_tags])

    for r in range(W.shape[0]):
        line = all_tags[r].ljust(pad + 1)

        if (W[r, :] == 0).all():
            line += "(!) "
        else:
            line += "    "

        for c in range(W.shape[1]):
            if skip_bit[c]:
                continue

            b = W[r, c]
            if b < 0:
                line += "0"
            elif b > 0:
                line += "1"
            else:
                line += "-"

        if tag_stats is not None:
            stat = tag_stats[all_tags[r]]
            line += " %4d|%4d" % stat

        x_min = np.min(X[r, :])
        x_max = np.max(X[r, :])
        line += " lo=%+.3f hi=%+.3f e=%.3f" % (x_min, x_max, E[r])

        lines.append(line)

    lines.append("")

    # Write
    for line in lines:
        fp.write(line + "\n")


def dump_solution_to_csv(fp, all_tags, all_bits, X):
    """
    Dumps solution data to CSV.

    Parameters
    ----------

    fp:
        An open file or stream
    all_tags:
        List of considered tags.
    all_bits:
        List of considered bits.
    X:
        Matrix with raw solution (floats).
    """

    # Bits
    line = ","
    for bit in all_bits:
        line += bit + ","
    fp.write(line[:-1] + "\n")

    # Tags + numbers
    for r, tag in enumerate(all_tags):
        line = tag + ","
        for c in range(X.shape[1]):
            line += "%+e," % X[r, c]
        fp.write(line[:-1] + "\n")


def dump_correlation_report(
        fp, all_tags, all_bits, W, C, correlation_exceptions):

    for i, tag in enumerate(all_tags):

        # No exceptions (100% correlation)
        if len(correlation_exceptions[tag]) == 0:
            continue

        fp.write(tag + "\n")

        for j, bit in enumerate(all_bits):

            if bit not in correlation_exceptions[tag]:
                continue

            c = C[i, j]
            w = W[i, j]

            # Dump bit correlation factor
            sgn = "+" if w > 0 else "-"
            fp.write(" bit %s: (%s) %.1f%%\n" % (bit.ljust(6), sgn, c * 100.0))

            # Dump counter-factual cases
            e = correlation_exceptions[tag][bit]
            for x, y, ex in e:
                fp.write("  is %d, should be %d - %s\n" % (x, y, ex))

    fp.write("\n")


# =============================================================================


def build_matrices(all_tags, all_bits, segdata, bias=0.0):
    """
    Builds matrices for the linear equation system to be solved.

    Parameters
    ----------

    all_tags:
        List of considered tags.
    all_bits:
        List of considered bits.
    segdata:
        List of segdata used.
    bias:
        T.B.D.
    """

    M = len(segdata)
    N = len(all_bits)
    K = len(all_tags)

    A = np.zeros((M, N), dtype=np.float64)
    B = np.zeros((M, K), dtype=np.float64)

    # A matrix
    for r, c in itertools.product(range(M), range(N)):
        if all_bits[c] in segdata[r]["bit"]:
            A[r, c] = +1.0
        else:
            A[r, c] = -1.0

    # B matrix
    for r, c in itertools.product(range(M), range(K)):
        for t, x in segdata[r]["tag"]:
            if t == all_tags[c]:
                v = +1.0 if x > 0 else -1.0
                B[r, c] = v + bias

    return A, B


def compute_error(A, B, X):
    """
    Computes solution error.

    Parameters
    ----------

    A:
        Matrix A
    B:
        Matrix B
    X:
        Matrix with computed solution.

    Returns
    -------

    A vector with errors
    """

    K = B.shape[1]

    # Compute error
    Bx = np.matmul(A, X)
    E = np.empty((K))
    for k in range(K):
        E[k] = np.sqrt(np.sum(np.square(Bx[:, k] - B[:, k])))

    return E


# =============================================================================


def solve_lms(all_tags, all_bits, segdata, bias=0.0):
    """
    Solves using direct least square solution (NumPy)

    Parameters
    ----------

    all_tags:
        List of considered tags.
    all_bits:
        List of considered bits.
    segdata:
        List of segdata used.
    bias:
        T.B.D.
    """

    # Build matrices
    A, B = build_matrices(all_tags, all_bits, segdata, bias)

    # Solve
    X, res, r, s = linalg.lstsq(A, B, rcond=None)

    return X, compute_error(A, B, X)


def solve_tichonov(all_tags, all_bits, segdata, bias=0.0, a=0.0):
    """
    Solves using Tichonov regularization method.

    Parameters
    ----------

    all_tags:
        List of considered tags.
    all_bits:
        List of considered bits.
    segdata:
        List of segdata used.
    bias:
        T.B.D.
    a:
        Regularization coefficient.

    Returns
    -------

    Tuple with:
    - Solution matrix X
    - Error vector.

    """

    M = len(segdata)
    N = len(all_bits)
    K = len(all_tags)

    # Build matrices
    A, B = build_matrices(all_tags, all_bits, segdata, bias)

    # Tikhonov regularization
    # https://en.wikipedia.org/wiki/Tikhonov_regularization
    AtA = np.matmul(A.T, A)
    AtB = np.matmul(A.T, B)
    X = np.matmul(np.linalg.inv(AtA + a * np.eye(N)), AtB)

    return X, compute_error(A, B, X)


# =============================================================================


def solve_onebyone(all_tags, all_bits, segdata, solver=solve_lms, **kw):
    """
    Solves each tag separately in one-by-one fashion.

    Parameters
    ----------

    all_tags:
        List of considered tags.
    all_bits:
        List of considered bits.
    segdata:
        List of segdata used.
    solver:
        Solver function.
    **kw:
        Parameters to solver function.

    Returns
    -------

    Tuple with:
    - Solution matrix X
    - Error vector.

    """

    X = np.empty((len(all_bits), len(all_tags)))
    E = np.empty((len(all_tags)))

    for i, tag in enumerate(all_tags):

        tag_segdata = [
            data for data in segdata if tag in [t[0] for t in data["tag"]]
        ]
        print("%s #%d" % (tag, len(tag_segdata)))

        X1, E1 = solver([tag], all_bits, tag_segdata, **kw)

        X[:, i] = X1[:, 0]
        E[i] = E1[0]

    return X, E


# =============================================================================


def detect_candidates(X, th, norm=None):
    """
    Detects candidate bits.

    Parameters
    ----------

    X:
        Matrix with solution
    th:
        Threshold
    norm:
        Normalization scheme. See code.

    Returns
    -------

    A tuple with:
    - Binary solution matrix W
    - Transposed matrix X

    """

    Xt = np.array(X.T)
    W = np.zeros_like(Xt, dtype=int)

    if norm == "max_abs":
        Nv = np.max(np.abs(Xt), axis=1)
        Xt /= np.tile(Nv[:, None], (1, Xt.shape[1]))

    W[Xt < -th] = -1
    W[Xt > +th] = +1

    return W, X.T


# =============================================================================


def compute_bit_correlations(tags_to_solve, bits_to_solve, segdata, W):
    """
    Basing on solution given in the matrix W returns a matrix C with
    correlation coefficients of each bit.

    Also returns a dict of dicts indexed by tag names and bit names with
    correlation exceptions - concrete specimen names where the correlation
    does not occur.
    """

    C = np.zeros_like(W, dtype=float)
    exceptions = {}

    for i, tag in enumerate(tags_to_solve):

        # Filter data for this tag
        tag_segdata = [
            data for data in segdata if tag in [t[0] for t in data["tag"]]
        ]
        exceptions[tag] = {}

        # Compute bit correlation
        for j, bit in enumerate(bits_to_solve):
            w = W[i, j]

            # No correlation with that bit
            if w == 0:
                continue

            corr_sum = 0
            corr_count = 0

            # Compute for one bit
            for k, data in enumerate(tag_segdata):
                bits = data["bit"]

                vt = [v for t, v in data["tag"] if t == tag][0]
                vb = 1 if bit in bits else 0

                # Negative correlation
                if w < 0:
                    vt = int(1 - vt)
                else:
                    vt = int(vt)

                # Correlates
                if vt == vb:
                    corr_sum += 1
                # Does not correlate
                else:
                    if bit not in exceptions[tag]:
                        exceptions[tag][bit] = []
                    exceptions[tag][bit].append((
                        vb,
                        vt,
                        data["seg"],
                    ))

                corr_count += 1

            # Store correlation
            C[i, j] = corr_sum / corr_count

    return C, exceptions


def compute_tag_stats(all_tags, segdata):
    """
    Counts occurrence of all considered tags

    Parameters
    ----------

    all_tags:
        Considered tags
    segdata:
        List of segdata used

    Returns
    -------

    A dict indexed by tag name with tuples containing 0 and 1 occurrence count.

    """

    stats = {}

    for i, tag in enumerate(all_tags):
        count0 = 0
        count1 = 0

        for data in segdata:
            for t, v in data["tag"]:
                if t == tag:
                    if v > 0:
                        count1 += 1
                    else:
                        count0 += 1

        stats[tag] = (
            count0,
            count1,
        )
    return stats


def sort_bits(bit_name):
    """
    Utility function for sorting bits.
    """

    frm, ofs = bit_name.split("_")
    return (
        int(frm),
        int(ofs),
    )


def build_address_map(tilegrid_file):
    """
    Loads the tilegrid and generates a map (baseaddr, offset) -> tile name(s).

    Parameters
    ----------

    tilegrid_file:
        The tilegrid.json file/

    Returns
    -------

    A dict with lists of tile names.

    """

    address_map = {}

    # Load tilegrid
    with OpenSafeFile(tilegrid_file, "r") as fp:
        tilegrid = json.load(fp)

    # Loop over tiles
    for tile_name, tile_data in tilegrid.items():

        # No bits or bits empty
        if "bits" not in tile_data:
            continue
        if not len(tile_data["bits"]):
            continue

        bits = tile_data["bits"]

        # No bus
        if "CLB_IO_CLK" not in bits:
            continue

        bus = bits["CLB_IO_CLK"]

        # Make the address as integers
        baseaddr = int(bus["baseaddr"], 16)
        offset = int(bus["offset"])
        address = (
            baseaddr,
            offset,
        )

        # Add tile to the map
        if address not in address_map:
            address_map[address] = []
        address_map[address].append(tile_name)

    return address_map


# =============================================================================


class FileOrStream(object):
    def __init__(self, file_name, stream=sys.stdout):
        self.file_name = file_name
        self.stream = stream
        self.fp = None

    def __enter__(self):
        if self.file_name is None:
            return self.stream
        if self.file_name == "-":
            return self.stream

        self.fp = open(self.file_name, "w")
        return self.fp

    def __exit__(self, exc_typ, exc_val, exc_tb):
        if self.fp is not None:
            self.fp.close()


# =============================================================================


def main():
    """
    The main.
    """

    # Parse arguments
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument(
        "files",
        nargs="*",
        type=str,
        help="Input file(s) generated by segmaker")
    parser.add_argument(
        "-o",
        type=str,
        default="segbits.rdb",
        help="Output database file (def. segbits.rdb)")
    parser.add_argument(
        "-f",
        type=str,
        default=None,
        help="Tag filter. Processes only tags containing the specified text")
    parser.add_argument(
        "-t", type=float, default=0.95, help="Candidate threshold (def. 0.95)")
    parser.add_argument(
        "-e",
        type=float,
        default=0.1,
        help="RMS error threshold below which a tag is rejected (def. 0.1)")
    parser.add_argument(
        "-a",
        type=float,
        default=0.01,
        help="Regularization coefficient (def. 0.01)")
    parser.add_argument(
        "--all",
        action="store_true",
        help="Solve all tags at once (may give worse results)")
    parser.add_argument(
        "-x",
        type=str,
        default=None,
        help="A CSV file name to Write the numerical solution to")
    parser.add_argument(
        "-r",
        type=str,
        default=None,
        help=
        "A text file name to write bit correlation report to. Specify '-' for stdout"
    )

    parser.add_argument(
        "-m",
        type=str,
        default=None,
        help="Mask bits found for this feature in all other features")

    parser.add_argument("-b", type=float, default=0.0, help="Bias")

    parser.add_argument("-no_0", action="store_true", help="Do not output 0s")
    parser.add_argument("-no_1", action="store_true", help="Do not output 1s")

    args = parser.parse_args()

    # Build (baseaddr, offset) -> tile name map
    database_dir = os.path.join(
        os.getenv("XRAY_DATABASE_DIR"), os.getenv("XRAY_DATABASE"),
        os.getenv("XRAY_FABRIC"))
    tilegrid_file = os.path.join(database_dir, "tilegrid.json")
    address_map = build_address_map(tilegrid_file)

    # Compute threshold
    th = args.t

    # Load and filter segdata
    segdata = []

    def tagfilter(tag):
        if args.f is None:
            return True
        return args.f in tag

    for name in args.files:
        print(name)
        segdata.extend(load_data(name, tagfilter, address_map))

    # Make list of all bits
    all_bits = set()
    for seg in segdata:
        all_bits |= set(seg["bit"])
    all_bits = sorted(list(all_bits), key=sort_bits)

    # Detect bits that are always set
    const1_bits = set(all_bits)
    for seg in segdata:
        const1_bits &= set(seg["bit"])

    # Make list of all tags
    all_tags = set()
    for seg in segdata:
        all_tags |= set([tag[0] for tag in seg["tag"]])
    all_tags = sorted(list(all_tags))

    # Count 0s and 1s for each tag
    tag_count = {}
    for seg in segdata:
        for tag, val in seg["tag"]:

            if tag not in tag_count:
                tag_count[tag] = [0, 0]

            if val > 0:
                tag_count[tag][1] += 1
            else:
                tag_count[tag][0] += 1

    # Identify const0 and const1 tags
    const_tags = {}
    for tag in all_tags:
        if tag_count[tag][0] == 0:
            const_tags[tag] = 1
        if tag_count[tag][1] == 0:
            const_tags[tag] = 0

    const0_tags = [t for t, v in const_tags.items() if v == 0]
    const1_tags = [t for t, v in const_tags.items() if v == 1]

    # Print config
    print("# segs:", len(segdata))
    print("# tags:", len(all_tags))
    print("# bits:", len(all_bits))
    print("threshold: %.2f" % th)

    if len(segdata) == 0:
        print("No data!")
        exit(-1)

    if len(all_tags) == 0:
        print("No tags!")
        exit(-1)

    if len(all_bits) == 0:
        print("No bits!")
        exit(-1)

    if len(const1_bits):
        print("const 1 bits: " + ", ".join(const1_bits))

    if len(const0_tags):
        print("const 0 tags: " + ", ".join(const0_tags))
    if len(const1_tags):
        print("const 1 tags: " + ", ".join(const1_tags))

    # Data to solve
    tags_to_solve = list(all_tags)
    bits_to_solve = list(all_bits)

    for tag in const_tags.keys():
        tags_to_solve.remove(tag)
    for bit in const1_bits:
        bits_to_solve.remove(bit)

    # Statistics
    tag_stats = compute_tag_stats(tags_to_solve, segdata)

    # Solve
    print("Solving...")
    if args.all:
        X, E = solve_tichonov(
            tags_to_solve, bits_to_solve, segdata, bias=args.b, a=args.a)
    else:
        X, E = solve_onebyone(
            tags_to_solve,
            bits_to_solve,
            segdata,
            solver=solve_tichonov,
            bias=args.b,
            a=args.a)

    # Detect candidate bits
    W, X = detect_candidates(X, th, norm="max_abs")

    # Mask
    if args.m is not None:
        print("Masking out %s" % args.m)
        tags = [t for t in tags_to_solve if args.m in t]
        for tag in tags:
            i = tags_to_solve.index(tag)
            for r in range(len(tags_to_solve)):
                if r == i:
                    continue
                for c in range(len(bits_to_solve)):
                    if W[r, c] == W[i, c]:
                        W[r, c] = 0

    # Reject 0s and/or 1s
    if args.no_0:
        W[W < 0] = 0
    if args.no_1:
        W[W > 0] = 0

    # Reject tags with error greater than threshold
    for r in range(X.shape[0]):
        if E[r] > args.e:
            W[r, :] = 0

    # Compute correlation
    C, correlation_exceptions = compute_bit_correlations(
        tags_to_solve, bits_to_solve, segdata, W)

    # Write segbits
    write_segbits(args.o, tags_to_solve, bits_to_solve, W)

    # Dump to CSV
    if args.x is not None:
        with OpenSafeFile(args.x, "w") as fp:
            dump_solution_to_csv(fp, tags_to_solve, bits_to_solve, X)

    # Dump results
    dump_results(sys.stdout, tags_to_solve, bits_to_solve, W, X, E, tag_stats)

    # Dump correlation report
    if args.r is not None:
        if args.r != "-":
            print("Dumping bit correlation report to '{}'".format(args.r))
        with FileOrStream(args.r, sys.stdout) as fp:
            dump_correlation_report(
                fp, tags_to_solve, bits_to_solve, W, C, correlation_exceptions)


# =============================================================================

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray.db import Database
import json


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        tile_type = tile_name.rsplit("_", 1)[0]

        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['PLLE2_ADV']:
                yield tile_name, tile_type, site_name


def main():
    sites = sorted(list(gen_sites()))
    max_sites = len(sites)

    f = open('params.jl', 'w')
    f.write('module,loc,params\n')

    routes_file = open('routes.txt', 'w')

    print(
        """
module top(
    input [{N}:0] clkin1,
    input [{N}:0] clkin2,
    input [{N}:0] clkfb,
    input [{N}:0] dclk
);

    (* KEEP, DONT_TOUCH *)
    LUT1 dummy();
""".format(N=max_sites - 1))

    for i, (
            tile_name,
            tile_type,
            site,
    ) in enumerate(sorted(gen_sites())):
        params = {
            "site":
            site,
            'active':
            random.random() > .2,
            "clkin1_conn":
            random.choice(
                ("clkfbout_mult_BUFG_" + site, "clkin1[{}]".format(i), "")),
            "clkin2_conn":
            random.choice(
                ("clkfbout_mult_BUFG_" + site, "clkin2[{}]".format(i), "")),
            "dclk_conn":
            random.choice((
                "0",
                "dclk[{}]".format(i),
            )),
            "dwe_conn":
            random.choice((
                "",
                "1",
                "0",
                "dwe_" + site,
                "den_" + site,
            )),
            "den_conn":
            random.choice((
                "",
                "1",
                "0",
                "den_" + site,
            )),
            "daddr4_conn":
            random.choice((
                "0",
                "dwe_" + site,
            )),
            "IS_RST_INVERTED":
            random.randint(0, 1),
            "IS_PWRDWN_INVERTED":
            random.randint(0, 1),
            "IS_CLKINSEL_INVERTED":
            random.randint(0, 1),
            "CLKFBOUT_MULT":
            random.randint(2, 4),
            "CLKOUT0_DIVIDE":
            random.randint(1, 128),
            "CLKOUT1_DIVIDE":
            random.randint(1, 128),
            "CLKOUT2_DIVIDE":
            random.randint(1, 128),
            "CLKOUT3_DIVIDE":
            random.randint(1, 128),
            "CLKOUT4_DIVIDE":
            random.randint(1, 128),
            "CLKOUT5_DIVIDE":
            random.randint(1, 128),
            "DIVCLK_DIVIDE":
            random.randint(1, 5),
            "CLKOUT0_DUTY_CYCLE":
            "0.500",
            "STARTUP_WAIT":
            verilog.quote('TRUE' if random.randint(0, 1) else 'FALSE'),
            "COMPENSATION":
            verilog.quote(
                random.choice((
                    'ZHOLD',
                    'BUF_IN',
                    'EXTERNAL',
                    'INTERNAL',
                ))),
            "BANDWIDTH":
            verilog.quote(random.choice((
                'OPTIMIZED',
                'HIGH',
                'LOW',
            ))),
        }

        if verilog.unquote(params['COMPENSATION']) == 'ZHOLD':
            params['clkfbin_conn'] = random.choice(
                (
                    "",
                    "clkfbout_mult_BUFG_" + site,
                ))
        elif verilog.unquote(params['COMPENSATION']) == 'INTERNAL':
            params['clkfbin_conn'] = random.choice(
                (
                    "",
                    "clkfbout_mult_" + site,
                ))
        else:
            params['clkfbin_conn'] = random.choice(
                ("", "clkfb[{}]".format(i), "clkfbout_mult_BUFG_" + site))

        params['clkin1_route'] = random.choice(
            (
                "{}_CLKIN1",
                "{}_FREQ_BB0",
                "{}_FREQ_BB1",
                "{}_FREQ_BB2",
                "{}_FREQ_BB3",
                "{}_PLLE2_CLK_IN1_INT",
            )).format(tile_type)

        params['clkin2_route'] = random.choice(
            (
                "{}_CLKIN2",
                "{}_FREQ_BB0",
                "{}_FREQ_BB1",
                "{}_FREQ_BB2",
                "{}_FREQ_BB3",
                "{}_PLLE2_CLK_IN2_INT",
            )).format(tile_type)

        params['clkfbin_route'] = random.choice(
            (
                "{}_CLKFBOUT2IN",
                "{}_UPPER_T_FREQ_BB0",
                "{}_UPPER_T_FREQ_BB1",
                "{}_UPPER_T_FREQ_BB2",
                "{}_UPPER_T_FREQ_BB3",
                "{}_UPPER_T_PLLE2_CLK_FB_INT",
            )).format(tile_type.replace("_UPPER_T", ""))

        f.write('%s\n' % (json.dumps(params)))

        def make_ibuf_net(net):
            p = net.find('[')
            return net[:p] + '_IBUF' + net[p:]

        if params['clkin1_conn'] != "":
            net = make_ibuf_net(params['clkin1_conn'])
            wire = '{}/{}'.format(tile_name, params['clkin1_route'])
            routes_file.write('{} {}\n'.format(net, wire))

        if params['clkin2_conn'] != "":
            net = make_ibuf_net(params['clkin2_conn'])
            wire = '{}/{}'.format(tile_name, params['clkin2_route'])
            routes_file.write('{} {}\n'.format(net, wire))

        if params['clkfbin_conn'] != "" and\
           params['clkfbin_conn'] != ("clkfbout_mult_BUFG_" + site):
            net = params['clkfbin_conn']
            if "[" in net and "]" in net:
                net = make_ibuf_net(net)
            wire = '{}/{}'.format(tile_name, params['clkfbin_route'])
            routes_file.write('{} {}\n'.format(net, wire))

        if not params['active']:
            continue

        print(
            """

    wire den_{site};
    wire dwe_{site};

    LUT1 den_lut_{site} (
        .O(den_{site})
    );

    LUT1 dwe_lut_{site} (
        .O(dwe_{site})
    );

    wire clkfbout_mult_{site};
    wire clkfbout_mult_BUFG_{site};
    wire clkout0_{site};
    wire clkout1_{site};
    wire clkout2_{site};
    wire clkout3_{site};
    wire clkout4_{site};
    wire clkout5_{site};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    PLLE2_ADV #(
            .IS_RST_INVERTED({IS_RST_INVERTED}),
            .IS_PWRDWN_INVERTED({IS_PWRDWN_INVERTED}),
            .IS_CLKINSEL_INVERTED({IS_CLKINSEL_INVERTED}),
            .CLKOUT0_DIVIDE({CLKOUT0_DIVIDE}),
            .CLKOUT1_DIVIDE({CLKOUT1_DIVIDE}),
            .CLKOUT2_DIVIDE({CLKOUT2_DIVIDE}),
            .CLKOUT3_DIVIDE({CLKOUT3_DIVIDE}),
            .CLKOUT4_DIVIDE({CLKOUT4_DIVIDE}),
            .CLKOUT5_DIVIDE({CLKOUT5_DIVIDE}),
            .CLKFBOUT_MULT({CLKFBOUT_MULT}),
            .DIVCLK_DIVIDE({DIVCLK_DIVIDE}),
            .STARTUP_WAIT({STARTUP_WAIT}),
            .CLKOUT0_DUTY_CYCLE({CLKOUT0_DUTY_CYCLE}),
            .COMPENSATION({COMPENSATION}),
            .BANDWIDTH({BANDWIDTH}),
            .CLKIN1_PERIOD(10.0),
            .CLKIN2_PERIOD(10.0)
    ) pll_{site} (
            .CLKFBOUT(clkfbout_mult_{site}),
            .CLKOUT0(clkout0_{site}),
            .CLKOUT1(clkout1_{site}),
            .CLKOUT2(clkout2_{site}),
            .CLKOUT3(clkout3_{site}),
            .CLKOUT4(clkout4_{site}),
            .CLKOUT5(clkout5_{site}),
            .DRDY(),
            .LOCKED(),
            .DO(),
            .CLKFBIN({clkfbin_conn}),
            .CLKIN1({clkin1_conn}),
            .CLKIN2({clkin2_conn}),
            .CLKINSEL(),
            .DCLK({dclk_conn}),
            .DEN({den_conn}),
            .DWE({dwe_conn}),
            .PWRDWN(),
            .RST(),
            .DI(),
            .DADDR({{7{{ {daddr4_conn} }} }}));

    (* KEEP, DONT_TOUCH *)
    BUFG bufg_{site} (
        .I(clkfbout_mult_{site}),
        .O(clkfbout_mult_BUFG_{site})
    );

    (* KEEP, DONT_TOUCH *)
    FDRE reg_clkfbout_mult_{site} (
        .C(clkfbout_mult_{site})
    );

    (* KEEP, DONT_TOUCH *)
    FDRE reg_clkout0_{site} (
        .C(clkout0_{site})
    );

    (* KEEP, DONT_TOUCH *)
    FDRE reg_clkout1_{site} (
        .C(clkout1_{site})
    );

    (* KEEP, DONT_TOUCH *)
    FDRE reg_clkout2_{site} (
        .C(clkout2_{site})
    );

    (* KEEP, DONT_TOUCH *)
    FDRE reg_clkout3_{site} (
        .C(clkout3_{site})
    );

    (* KEEP, DONT_TOUCH *)
    FDRE reg_clkout4_{site} (
        .C(clkout4_{site})
    );

    (* KEEP, DONT_TOUCH *)
    FDRE reg_clkout5_{site} (
        .C(clkout5_{site})
    );
            """.format(**params))

    print('endmodule')

    f.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import re
import os

from prjxray.segmaker import Segmaker
from prjxray.bitfilter import get_bitfilter

segmk = Segmaker("design.bits")

tiledata = dict()
pipdata = dict()
ignpip = set()

print("Loading tags from design.txt.")
with open("design.txt", "r") as f:
    for line in f:
        tile, pip, src, dst, pnum, pdir = line.split()

        if not tile.startswith('INT_'):
            continue

        _, pip = pip.split(".")
        _, src = src.split("/")
        _, dst = dst.split("/")
        pnum = int(pnum)
        pdir = int(pdir)

        if tile not in tiledata:
            tiledata[tile] = {"pips": set(), "srcs": set(), "dsts": set()}

        if pip in pipdata:
            assert pipdata[pip] == (src, dst)
        else:
            pipdata[pip] = (src, dst)

        tiledata[tile]["pips"].add(pip)
        tiledata[tile]["srcs"].add(src)
        tiledata[tile]["dsts"].add(dst)

        if pdir == 0:
            tiledata[tile]["srcs"].add(dst)
            tiledata[tile]["dsts"].add(src)

        gfan_src = re.match('^GFAN', src) is not None
        fan_alt_byp_bounce = re.match('^FAN_ALT[0-9]',
                                      dst) is not None and re.match(
                                          '^BYP_BOUNCE.*', src) is not None
        byp_alt_imux_dst = re.match(
            '^(BYP_ALT[0-9]+|IMUX_?L?[0-9]+)', dst) is not None
        logic_out_src = re.match('^LOGIC_OUTS_?L?[0-9]+', src) is not None

        # Okay: BYP_ALT0.VCC_WIRE
        # Skip: INT.IMUX13.VCC_WIRE, INT.IMUX_L43.VCC_WIRE
        # Skip: INT.BYP_ALT[0-9].LOGIC_OUTS_ and INT.IMUX_?L?. LOGIC_OUTS_
        if pnum == 1 or pdir == 0 or \
                src == "VCC_WIRE" or \
                re.match(r"^(L[HV]B?|G?CLK)(_L)?(_B)?[0-9]", src) or \
                re.match(r"^(L[HV]B?|G?CLK)(_L)?(_B)?[0-9]", dst) or \
                gfan_src or \
                fan_alt_byp_bounce or \
                (logic_out_src and byp_alt_imux_dst) or \
                re.match(r"^(CTRL|GFAN)(_L)?[0-9]", dst):
            ignpip.add(pip)

for tile, pips_srcs_dsts in tiledata.items():
    pips = pips_srcs_dsts["pips"]
    srcs = pips_srcs_dsts["srcs"]
    dsts = pips_srcs_dsts["dsts"]

    for pip, src_dst in pipdata.items():
        src, dst = src_dst
        if pip in ignpip:
            pass
        elif pip in pips:
            segmk.add_tile_tag(tile, "%s.%s" % (dst, src), 1)
        elif src_dst[1] not in dsts:
            segmk.add_tile_tag(tile, "%s.%s" % (dst, src), 0)

segmk.compile(bitfilter=get_bitfilter(os.getenv('XRAY_DATABASE'), 'INT'))
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Sanity checks FASM output from IOB fuzzer.
The IOB fuzzer is fairly complicated, and it's output is hard to verify by
inspected.  For this reason, check_results.py was written to compare the
specimen's generated and their FASM output.  The FASM output does pose a
chicken and egg issue.  The test procedure is a follows:

1. Build the database (e.g. make -j<N> run)
2. Build the database again (e.g. make -j<N> run)
3. Run check_results.py

The second time that the database is run, the FASM files in the specimen's
will have the bits documented by fuzzer.

"""
import argparse
import os
import os.path
from prjxray import verilog
import json
import generate


def process_parts(parts):
    if len(parts) == 0:
        return

    if parts[-1] == 'IN_ONLY':
        yield 'type', ['IBUF', 'IBUFDS']

    if len(parts) > 2 and parts[-2] == 'SLEW':
        yield 'SLEW', verilog.quote(parts[-1])

    if parts[0] == 'PULLTYPE':
        yield 'PULLTYPE', verilog.quote(parts[1])

    if len(parts) > 1 and parts[1] == 'IN':
        yield 'IOSTANDARDS', parts[0].split('_')
        yield 'IN', True

    if len(parts) > 1 and parts[1] == 'IN_DIFF':
        yield 'IOSTANDARDS', parts[0].split('_')
        yield 'IN_DIFF', True

    if len(parts) > 1 and parts[1] == 'DRIVE':
        yield 'IOSTANDARDS', parts[0].split('_')

        if parts[2] == 'I_FIXED':
            yield 'DRIVES', [None]
        else:
            yield 'DRIVES', parts[2].split('_')


def create_sites_from_fasm(fasm_file):
    sites = {}

    diff_tiles = set()

    with open(fasm_file) as f:
        for l in f:
            if 'IOB33' not in l:
                continue

            parts = l.strip().split('.')
            tile = parts[0]
            site = parts[1]

            if 'OUT_DIFF' == site:
                diff_tiles.add(tile)
                continue

            if (tile, site) not in sites:
                sites[(tile, site)] = {
                    'tile': tile,
                    'site_key': site,
                }

            if len(parts) > 3 and 'IN_DIFF' == parts[3]:
                diff_tiles.add(tile)

            for key, value in process_parts(parts[2:]):
                sites[(tile, site)][key] = value

    for key in sites:
        if 'type' not in sites[key]:
            if 'IOSTANDARDS' not in sites[key]:
                sites[key]['type'] = [None]
            else:
                assert 'IOSTANDARDS' in sites[key], sites[key]
                assert 'DRIVES' in sites[key], sites[key]

                if 'IN' in sites[key]:
                    sites[key]['type'] = ['IOBUF', 'IOBUF_INTERMDISABLE']
                else:
                    sites[key]['type'] = [
                        "OBUF",
                        "OBUFDS",
                        "OBUFTDS",
                        "OBUFDS_DUAL_BUF",
                        "OBUFTDS_DUAL_BUF",
                    ]

    return sites, diff_tiles


def process_specimen(fasm_file, params_json):
    sites, diff_tiles = create_sites_from_fasm(fasm_file)

    with open(params_json) as f:
        params = json.load(f)

    count = 0
    for p in params['tiles']:
        tile = p['tile']
        for site in p['site'].split(' '):
            site_y = int(site[site.find('Y') + 1:]) % 2

            if generate.skip_broken_tiles(p):
                continue

            site_key = 'IOB_Y{}'.format(site_y)

            if (tile, site_key) not in sites:
                assert p['type'] is None, p
                continue

            site_from_fasm = sites[(tile, site_key)]

            if site_y == 0 or tile not in diff_tiles:
                assert p['type'] in site_from_fasm['type'], (
                    tile, site_key, p['type'], site_from_fasm['type'])
            else:
                # Y1 on DIFF tiles is always none.
                assert p['type'] is None, p

            if p['type'] is None:
                continue

            assert 'PULLTYPE' in p, p
            assert 'PULLTYPE' in site_from_fasm, site_from_fasm

            if verilog.unquote(p['PULLTYPE']) == '':
                # Default is None.
                pulltype = verilog.quote('NONE')
            else:
                pulltype = p['PULLTYPE']

            assert pulltype == site_from_fasm['PULLTYPE'], (
                tile, site_key, p, site_from_fasm)

            assert 'IOSTANDARDS' in site_from_fasm, (tile, site)

            iostandard = verilog.unquote(p['IOSTANDARD'])
            if iostandard.startswith('DIFF_'):
                iostandard = iostandard[5:]

            assert iostandard in site_from_fasm['IOSTANDARDS'], (
                p['IOSTANDARD'],
                site_from_fasm['IOSTANDARDS'],
            )

            if p['type'] not in ['IBUF', 'IBUFDS']:
                if verilog.unquote(p['SLEW']) == '':
                    # Default is None.
                    slew = verilog.quote('SLOW')
                else:
                    slew = p['SLEW']

                assert slew == site_from_fasm['SLEW'], (
                    tile, site_key, p, site_from_fasm)

                assert 'DRIVES' not in p, p
                assert 'DRIVES' in site_from_fasm, (
                    tile, site, p['type'], site_from_fasm)

                if p['DRIVE'] is None:
                    assert None in site_from_fasm['DRIVES'], (
                        tile, site_key, p['DRIVE'], site_from_fasm['DRIVES'])
                elif p['DRIVE'] is '':
                    if None in site_from_fasm['DRIVES']:
                        # IOSTANDARD has not DRIVE setting, ignore
                        pass
                    else:
                        # Check that drive is at default
                        assert 'I12' in site_from_fasm['DRIVES'], (
                            tile, site_key, p['DRIVE'],
                            site_from_fasm['DRIVES'])
                else:
                    assert 'I{}'.format(
                        p['DRIVE']) in site_from_fasm['DRIVES'], (
                            tile, site_key, p['DRIVE'],
                            site_from_fasm['DRIVES'])

            count += 1

    return count


def scan_specimens():
    for root, dirs, files in os.walk('build'):
        if os.path.basename(root).startswith('specimen_'):
            print('Processing', os.path.basename(root))
            process_specimen(
                fasm_file=os.path.join(root, 'design.fasm'),
                params_json=os.path.join(root, 'params.json'))

    print('No errors found!')


def main():
    parser = argparse.ArgumentParser(description="Verify IOB FASM vs BELs.")

    parser.add_argument('--fasm')
    parser.add_argument('--params')

    args = parser.parse_args()

    if not args.fasm and not args.params:
        scan_specimens()
    else:
        count = process_specimen(fasm_file=args.fasm, params_json=args.params)
        print('No errors found in {} IO sites'.format(count))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker
import re

segmk = Segmaker("design.bits")

tiledata = {}
pipdata = {}
ignpip = set()

print("Loading tags from design.txt.")
with open("design.txt", "r") as f:
    for line in f:
        tile, pip, src, dst, pnum, pdir = line.split()

        if not tile.startswith('BRAM_'):
            continue

        # BRAM_R_X37Y0/BRAM_R.BRAM_IMUX35_1->BRAM_R_IMUX_ADDRBWRADDRL2
        pip_prefix, pip = pip.split(".")
        tile_from_pip, tile_type = pip_prefix.split('/')
        assert tile == tile_from_pip
        _, src = src.split("/")
        _, dst = dst.split("/")
        pnum = int(pnum)
        pdir = int(pdir)

        if tile not in tiledata:
            tiledata[tile] = {
                "type": tile_type,
                "pips": set(),
                "srcs": set(),
                "dsts": set()
            }

        if tile_type not in pipdata:
            pipdata[tile_type] = {}

        if pip in pipdata[tile_type]:
            assert pipdata[tile_type][pip] == (src, dst)
        else:
            pipdata[tile_type][pip] = (src, dst)

        tiledata[tile]["pips"].add(pip)
        tiledata[tile]["srcs"].add(src)
        tiledata[tile]["dsts"].add(dst)

        if pdir == 0:
            tiledata[tile]["srcs"].add(dst)
            tiledata[tile]["dsts"].add(src)

        if pnum == 1 or pdir == 0:
            ignpip.add(pip)

CASCOUT_RE = re.compile('^BRAM_CASCOUT_ADDR((?:BWR)|(?:ARD))ADDRU([0-9]+)$')

for tile, pips_srcs_dsts in tiledata.items():
    tile_type = pips_srcs_dsts["type"]
    pips = pips_srcs_dsts["pips"]
    srcs = pips_srcs_dsts["srcs"]
    dsts = pips_srcs_dsts["dsts"]

    active_cascout = set()
    for pip, src_dst in pipdata[tile_type].items():
        src, dst = src_dst

        # BRAM_R has some _R_ added to some pips.  Because BRAM_L and BRAM_R
        # appears to shares all bits, overlap the names during fuzzing to avoid
        # extra work.
        #
        # BRAM.BRAM_ADDRARDADDRL0.BRAM_IMUX_R_ADDRARDADDRL0
        #
        # becomes
        #
        # BRAM.BRAM_ADDRARDADDRL0.BRAM_IMUX_ADDRARDADDRL0
        src_no_r = src.replace('BRAM_R_IMUX_ADDR', 'BRAM_IMUX_ADDR')

        if pip in ignpip:
            pass
        elif pip in pips:
            segmk.add_tile_tag(tile, "%s.%s" % (dst, src_no_r), 1)
        elif src_dst[1] not in dsts:
            segmk.add_tile_tag(tile, "%s.%s" % (dst, src_no_r), 0)

        m = CASCOUT_RE.match(dst)
        if m and pip in pips:
            active_cascout.add(m.group(1))

    for group in ['BWR', 'ARD']:
        segmk.add_tile_tag(
            tile, 'CASCOUT_{}_ACTIVE'.format(group), group in active_cascout)

segmk.compile()
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import io
import json
import re

from collections import OrderedDict


def extract_numbers(s):
    """
    >>> extract_numbers("CLK_HROW_WR10END2_3")
    ('CLK_HROW_WR', 10, 'END', 2, '_', 3)
    >>> extract_numbers("VBRK_WR1END2")
    ('VBRK_WR', 1, 'END', 2)
    """
    bits = []
    for m in re.finditer("([^0-9]*)([0-9]*)", s):
        if m.group(1):
            bits.append(m.group(1))
        if m.group(2):
            bits.append(int(m.group(2)))
    return tuple(bits)


def sort(data):
    """Sort data types via "natural" numbers.

    Supports all the basic Python data types.
    >>> o = sort({
    ...    't1': {'c','b'},          # Set
    ...    't2': ('a2', 'a10', 'e'), # Tuple
    ...    't3': [5, 3, 2],          # List
    ...    't4': {                   # Dictionary
    ...        'a4': ('b2', 'b3'),
    ...        'a2': ['c1', 'c2', 'c0', 'c10'],
    ...    },
    ...    't5': ['a1b5', 'a2b1', 'a1b1'],
    ... })
    >>> for t in o:
    ...    print(t+':', o[t])
    t1: ('b', 'c')
    t2: ('a2', 'a10', 'e')
    t3: (5, 3, 2)
    t4: OrderedDict([('a2', ('c1', 'c2', 'c0', 'c10')), ('a4', ('b2', 'b3'))])
    t5: ('a1b5', 'a2b1', 'a1b1')

    Don't mangle "pairs"
    >>> sort([('b', 'c'), ('2', '1')])
    (('b', 'c'), ('2', '1'))
    """

    def key(o):
        if o is None:
            return None
        elif isinstance(o, str):
            return extract_numbers(o)
        elif isinstance(o, int):
            return o
        elif isinstance(o, (list, tuple)):
            return tuple(key(i) for i in o)
        elif isinstance(o, dict):
            return tuple((key(k), key(v)) for k, v in o.items())
        elif isinstance(o, set):
            return tuple(key(k) for k in o)
        raise ValueError(repr(o))

    def rsorter(o):
        if isinstance(o, dict):
            nitems = []
            for k, v in o.items():
                nitems.append((key(k), k, rsorter(v)))
            nitems.sort(key=lambda n: n[0])

            new_dict = OrderedDict()
            for _, k, v in nitems:
                new_dict[k] = v
            return new_dict

        elif isinstance(o, set):
            return tuple(sorted((rsorter(v) for v in o), key=key))
        elif isinstance(o, (tuple, list)):
            return tuple(rsorter(v) for v in o)
        else:
            return o

    return rsorter(data)


def pprint(f, data):
    detach = False
    if not isinstance(f, io.TextIOBase):
        detach = True
        f = io.TextIOWrapper(f)
    data = sort(data)
    json.dump(data, f, indent=4)
    f.write('\n')
    f.flush()
    if detach:
        f.detach()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json

from prjxray.segmaker import Segmaker


#Decouple interconnect bits from PRESELECT property bits
def bitfilter(frame, bit):
    if frame == 7 or frame == 21:
        return False
    return True


def main():
    segmk = Segmaker("design.bits")

    print("Loading tags")
    with open('params.json') as f:
        params = json.load(f)

    for row in params:
        base_name = 'BUFGCTRL_X{}Y{}'.format(row['x'], row['y'])

        segmk.add_site_tag(
            row['site'], '{}.IN_USE'.format(base_name), row['IN_USE'])

        if not row['IN_USE']:
            continue

        for param in (
                'INIT_OUT',
                'IS_IGNORE0_INVERTED',
                'IS_IGNORE1_INVERTED',
        ):
            segmk.add_site_tag(
                row['site'], '{}.{}'.format(base_name, param), row[param])

        for param in ('PRESELECT_I0', ):
            segmk.add_site_tag(
                row['site'], '{}.Z{}'.format(base_name, param), 1 ^ row[param])

        for param in ('PRESELECT_I1', ):
            segmk.add_site_tag(
                row['site'], '{}.{}'.format(base_name, param), row[param])

        for param, tag in (('IS_CE0_INVERTED', 'ZINV_CE0'), ('IS_S0_INVERTED',
                                                             'ZINV_S0'),
                           ('IS_CE1_INVERTED', 'ZINV_CE1'), ('IS_S1_INVERTED',
                                                             'ZINV_S1')):
            segmk.add_site_tag(
                row['site'], '{}.{}'.format(base_name, tag), 1 ^ row[param])

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker

segmk = Segmaker("design.bits")

print("Loading tags")
f = open('params.csv', 'r')
f.readline()
for l in f:
    module, loc, loc2 = l.split(',')

    tmp = module.replace('clb_PRECYINIT_0', 'C0')
    tmp = tmp.replace('clb_PRECYINIT_1', 'C1')
    # clb_PRECYINIT_AX => AX
    src = tmp.replace('clb_PRECYINIT_', '')
    '''
    PRECYINIT
                00_12   30_14   30_13
    C1          0       1       0
    AX          1       0       0
    CIN         0       0       1
    C0          0       0       0
    '''
    srcs = ('C0', 'C1', 'AX', 'CIN')
    for asrc in srcs:
        segmk.add_site_tag(loc, "PRECYINIT.%s" % asrc, int(src == asrc))

segmk.compile()
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json
from prjxray.segmaker import Segmaker

BITS_PER_PARAM = 256
NUM_INITP_PARAMS = 8
NUM_INIT_PARAMS = 0x40
BITS_PER_SITE = BITS_PER_PARAM * (NUM_INITP_PARAMS + NUM_INIT_PARAMS)


def main():
    segmk = Segmaker("design.bits")
    segmk.set_def_bt('BLOCK_RAM')

    print("Loading tags")
    '''
    '''

    with open('params.json') as f:
        params = json.load(f)

    for param in params:
        for initp in range(NUM_INITP_PARAMS):
            p = 'INITP_{:02X}'.format(initp)
            val = param[p]
            for bit in range(BITS_PER_PARAM):
                segmk.add_site_tag(
                    param['site'], "{p}[{bit:03d}]".format(
                        p=p,
                        bit=bit,
                    ), val & (1 << bit) != 0)

        for init in range(NUM_INIT_PARAMS):
            p = 'INIT_{:02X}'.format(init)
            val = param[p]
            for bit in range(BITS_PER_PARAM):
                segmk.add_site_tag(
                    param['site'], "{p}[{bit:03d}]".format(
                        p=p,
                        bit=bit,
                    ), val & (1 << bit) != 0)

    segmk.compile()
    segmk.write()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os, random
random.seed(0)
from prjxray import util
from prjxray import verilog

# INCREMENT is the amount of additional CLBN to be instantiated in the design.
# This makes the fuzzer compilation more robust against failures.
INCREMENT = os.getenv('CLBN', 0)
CLBN = 50 + int(INCREMENT)
print('//Requested CLBs: %s' % str(CLBN))


def gen_slicems():
    for _tile_name, site_name, _site_type in util.get_roi().gen_sites(
        ['SLICEM']):
        yield site_name


DIN_N = CLBN * 8
DOUT_N = CLBN * 8

verilog.top_harness(DIN_N, DOUT_N)

f = open('params.csv', 'w')
f.write('module,loc,c31,b31,a31\n')
slices = gen_slicems()
print(
    'module roi(input clk, input [%d:0] din, output [%d:0] dout);' %
    (DIN_N - 1, DOUT_N - 1))
multis = 0
for clbi in range(CLBN):
    loc = next(slices)
    module = 'my_NDI1MUX_NI_NMC31'
    c31 = random.randint(0, 1)
    b31 = random.randint(0, 1)
    a31 = random.randint(0, 1)

    print('    %s' % module)
    print(
        '           #(.LOC("%s"), .C31(%d), .B31(%d), .A31(%d))' %
        (loc, c31, b31, a31))
    print(
        '            clb_%d (.clk(clk), .din(din[  %d +: 8]), .dout(dout[  %d +: 8]));'
        % (clbi, 8 * clbi, 8 * clbi))

    f.write('%s,%s,%d,%d,%d\n' % (module, loc, c31, b31, a31))
f.close()
print(
    '''endmodule

// ---------------------------------------------------------------------

''')

print(
    '''
module my_NDI1MUX_NI_NMC31 (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC = "SLICE_X6Y100";
    parameter C31 = 0;
    parameter B31 = 0;
    parameter A31 = 0;

    wire [3:0] q31;

    wire [3:0] lutd;
    assign lutd[3] = din[7];
    assign lutd[2] = C31 ? q31[3] : din[7];
    assign lutd[1] = B31 ? q31[2] : din[7];
    assign lutd[0] = A31 ? q31[1] : din[7];

    (* LOC=LOC, BEL="D6LUT" *)
    SRLC32E #(
            .INIT(32'h00000000),
            .IS_CLK_INVERTED(1'b0)
        ) lutd (
            .Q(dout[0]),
            .Q31(q31[3]),
            .A(din[4:0]),
            .CE(din[5]),
            .CLK(din[6]),
            .D(lutd[3]));
    (* LOC=LOC, BEL="C6LUT" *)
    SRLC32E #(
            .INIT(32'h00000000),
            .IS_CLK_INVERTED(1'b0)
        ) lutc (
            .Q(dout[1]),
            .Q31(q31[2]),
            .A(din[4:0]),
            .CE(din[5]),
            .CLK(din[6]),
            .D(lutd[2]));
    (* LOC=LOC, BEL="B6LUT" *)
    SRLC32E #(
            .INIT(32'h00000000),
            .IS_CLK_INVERTED(1'b0)
        ) lutb (
            .Q(dout[2]),
            .Q31(q31[1]),
            .A(din[4:0]),
            .CE(din[5]),
            .CLK(din[6]),
            .D(lutd[1]));
    (* LOC=LOC, BEL="A6LUT" *)
    SRLC32E #(
            .INIT(32'h00000000),
            .IS_CLK_INVERTED(1'b0)
        ) luta (
            .Q(dout[3]),
            .Q31(q31[0]),
            .A(din[4:0]),
            .CE(din[5]),
            .CLK(din[6]),
            .D(lutd[0]));
endmodule
''')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

ports = {
    "DSP48E1": [
        ("ALUMODE", 4),
        ("CARRYINSEL", 3),
        ("CEAD", 1),
        ("CEALUMODE", 1),
        ("CED", 1),
        ("CEINMODE", 1),
        ("D", 25),
        ("INMODE", 5),
        ("OPMODE", 7),
        ("RSTD", 1),
    ],
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
from collections import namedtuple
import enum


class BlockType(enum.Enum):
    # Frames describing CLB features, interconnect, clocks and IOs.
    CLB_IO_CLK = 'CLB_IO_CLK'

    # Frames describing block RAM initialization.
    BLOCK_RAM = 'BLOCK_RAM'


GridLoc = namedtuple('GridLoc', 'grid_x grid_y')
ClockRegion = namedtuple('ClockRegion', 'name x y')
GridInfo = namedtuple(
    'GridInfo',
    'bits sites prohibited_sites tile_type pin_functions clock_region')
BitAlias = namedtuple('BitAlias', 'tile_type start_offset sites')
Bits = namedtuple('Bits', 'base_address frames offset words alias')
BitsInfo = namedtuple('BitsInfo', 'block_type tile bits')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from datetime import datetime, timedelta

import argparse
import collections
import io
import os
import os.path
import re
import resource
import signal
import subprocess
import sys
import traceback
from utils.create_environment import get_environment_variables

import junit_xml

secs_in_min = 60
secs_in_hour = 60 * secs_in_min


def pretty_timedelta_str(d):
    """Pretty print a time delta object.

    >>> pretty_timedelta_str(timedelta(seconds=2))
    '2s'

    >>> pretty_timedelta_str(timedelta(seconds=61))
    '1m01s'

    >>> pretty_timedelta_str(timedelta(seconds=125))
    '2m05s'

    >>> pretty_timedelta_str(timedelta(seconds=secs_in_hour * 5))
    '5h00m'

    Once we get to displaying hours, we don't bother displaying seconds.
    >>> pretty_timedelta_str(timedelta(seconds=secs_in_hour * 5 + 1))
    '5h00m'

    >>> pretty_timedelta_str(timedelta(
    ...   seconds=secs_in_hour * 5 + secs_in_min * 2 + 1))
    '5h02m'

    >>> pretty_timedelta_str(timedelta(seconds=secs_in_hour * 5.5))
    '5h30m'
    """
    assert isinstance(d, timedelta)
    ts = d.total_seconds()
    assert ts > 0, (d, ts)

    bits = []
    hours = int(ts / secs_in_hour)
    if hours > 0:
        bits.append("{}h".format(hours))
        ts -= hours * secs_in_hour

    mins = int(ts / secs_in_min)
    if mins > 0 or hours > 0:
        if hours > 0:
            bits.append("{:02d}m".format(mins))
        else:
            bits.append("{:d}m".format(mins))
        ts -= mins * secs_in_min

    if hours == 0:
        if mins > 0:
            bits.append("{:02d}s".format(int(ts)))
        else:
            bits.append("{:d}s".format(int(ts)))

    assert len(bits) > 0, d

    return "".join(bits)


class Logger:
    """Output a bunch of lines with a prefix.

    >>> l = Logger("fuzz", datetime(2001, 11, 1), 6)
    >>> l._now = lambda: datetime(2001, 11, 1, second=10)
    >>> l.log("Test!")
    2001-11-01T00:00:10 - fuzz   -   10s: Test!
    >>> l.log("Format {} {t}", [1,], {'t': 2})
    2001-11-01T00:00:10 - fuzz   -   10s: Format 1 2
    >>> l.log('''\\
    ... Line 1
    ... Line 2
    ... Line 3
    ... ''')
    2001-11-01T00:00:10 - fuzz   -   10s: Line 1
    2001-11-01T00:00:10 - fuzz   -   10s: Line 2
    2001-11-01T00:00:10 - fuzz   -   10s: Line 3
    """

    def __init__(self, fuzzer, time_start, padding):
        self.fuzzer = fuzzer + ' ' * (padding - len(fuzzer))
        self.time_start = time_start
        self.queue = collections.deque()

    def _now(self):
        return datetime.utcnow()

    def log(self, msg, args=None, kw=None, flush=True):
        """Queue a log message.

        This is safe to do in a signal handler where you can't do io.
        """
        if args is None:
            args = []
        if kw is None:
            kw = {}

        time_log = self._now()
        self.queue.append((time_log, msg, args, kw))
        if flush:
            self.flush()

    def flush(self):
        while len(self.queue) > 0:
            time_log, msg, args, kw = self.queue.popleft()
            self._output(time_log, msg, args, kw)

    def _output(self, time_log, msg, args, kw):
        running_for = time_log - self.time_start
        msg = msg.format(*args, **kw)

        time_log = time_log.replace(microsecond=0)

        log_prefix = "{:s} - {}/{:s} - {:>5s}: ".format(
            time_log.isoformat(),
            os.environ['XRAY_PART'],
            self.fuzzer,
            pretty_timedelta_str(running_for),
        )

        msg = "\n".join(log_prefix + x for x in msg.splitlines())
        print(msg, flush=True)


def last_lines(f, num, blocksize=100):
    r"""Return n last lines in a file.

    >>> f = io.StringIO("")
    >>> last_lines(f, 100)
    ['']

    >>> f = io.StringIO("1\n2")
    >>> last_lines(f, 100)
    ['1', '2']

    >>> f = io.StringIO("11\n22\n33\n44\n55")
    >>> last_lines(f, 100)
    ['11', '22', '33', '44', '55']

    >>> f = io.StringIO("11\n22\n33\n44\n55")
    >>> last_lines(f, 100, 2)
    ['11', '22', '33', '44', '55']

    >>> f = io.StringIO("11\n22\n33\n44\n55")
    >>> last_lines(f, 100, 3)
    ['11', '22', '33', '44', '55']

    >>> f = io.StringIO("11\n22\n33\n44\n55")
    >>> last_lines(f, 2, 2)
    ['44', '55']

    >>> f = io.StringIO("11\n22\n33\n44\n55")
    >>> last_lines(f, 2, 3)
    ['44', '55']

    >>> f = io.StringIO("11\n22\n33\n44\n55")
    >>> last_lines(f, 2, 4)
    ['44', '55']

    >>> f = io.StringIO("1\n22\n3333333\n")
    >>> last_lines(f, 100, 3)
    ['1', '22', '3333333', '']

    """
    lines = collections.deque([''])
    # Seek to the end of the file
    f.seek(0, os.SEEK_END)
    fpos = f.tell()

    while len(lines) < (num + 1) and fpos > 0:
        bs = min(blocksize, fpos)
        fpos = fpos - bs
        f.seek(fpos)
        data = f.read(bs)

        while True:
            lpos = data.rfind('\n')
            lines[0] = data[lpos + 1:] + lines[0]
            if lpos == -1:
                break
            lines.insert(0, '')
            data = data[:lpos]

    if fpos == 0:
        lines.insert(0, '')

    return list(lines)[1:]


def get_usage():
    # This function only works if you have a signal handler for the
    # signal.SIGCHLD signal.
    raw_usage = resource.getrusage(resource.RUSAGE_CHILDREN)
    # 0   ru_utime    time in user mode (float)
    # 1   ru_stime    time in system mode (float)
    # 2   ru_maxrss   maximum resident set size
    #
    # These fields are always zero on Linux
    # 3   ru_ixrss    shared memory size
    # 4   ru_idrss    unshared memory size
    # 5   ru_isrss    unshared stack size
    return "User:{}s System:{}s".format(
        int(raw_usage.ru_utime),
        int(raw_usage.ru_stime),
    )


def get_load():
    """Return current load average.

    Values is average perecntage of CPU used over 1 minute, 5 minutes, 15
    minutes.
    """
    a1, a5, a15 = os.getloadavg()
    cpus = os.cpu_count()
    return a1 / cpus * 100.0, a5 / cpus * 100.0, a15 / cpus * 100.0


class PsTree:
    _pstree_features = None

    @classmethod
    def get_features(cls):
        if cls._pstree_features is None:
            cls._pstree_features = []
            for f in ['-T', '-l']:
                try:
                    subprocess.check_output(
                        "pstree {}".format(f),
                        stderr=subprocess.STDOUT,
                        shell=True,
                    )
                    cls._pstree_features.append(f)
                except subprocess.CalledProcessError:
                    continue
        return cls._pstree_features

    @classmethod
    def get(cls, pid):
        """Get processes under current one.

        Requires the pstree until be installed, otherwise returns empty string.
        """
        p = subprocess.Popen(
            "pstree {} {}".format(" ".join(cls.get_features()), pid),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.DEVNULL,
            universal_newlines=True,
        )
        stdout, stderr = p.communicate()

        # Special case segfaults.
        if 'Segmentation fault' in stderr:
            stderr = ''
            stdout = '(result too big, pstree segfaulted.)'

        # If no error, just return stdout
        if not stderr:
            return stdout

        return "{}\n{}".format(stdout, stderr)


def mem_convert(s):
    """
    >>> mem_convert('62G')
    62000000000.0
    >>> mem_convert('62M')
    62000000.0
    >>> mem_convert('62B')
    62.0
    """
    units = {
        'Gi': pow(2, 30),
        'Mi': pow(2, 20),
        'Ki': pow(2, 10),
        'G': 1e9,
        'M': 1e6,
        'K': 1e3,
        'B': 1,
    }
    u = '',
    m = 1
    for u, m in units.items():
        if s.endswith(u):
            break

    v = float(s[:-len(u)])
    v = v * m
    return v


def get_memory(memstr=None):
    r"""
    >>> import pprint
    >>> pprint.pprint(get_memory('''\
    ...               total        used        free      shared  buff/cache   available
    ... Mem:           62G        19G       4.8G       661M        38G        42G
    ... Swap:           0B         0B         0B
    ... '''))
    {'mem': {'available': 42000000000.0,
             'buff/cache': 38000000000.0,
             'free': 4800000000.0,
             'shared': 661000000.0,
             'total': 62000000000.0,
             'used': 19000000000.0},
     'swap': {'free': 0.0, 'total': 0.0, 'used': 0.0}}
    """
    if memstr is None:
        memstr = subprocess.check_output(
            'free -mh', shell=True).decode("utf-8")

    lines = [x.split() for x in memstr.strip().splitlines()]
    lines[0].insert(0, 'type:')

    for l in lines:
        l[0] = l[0][:-1].lower()

    headers = lines[0][1:]
    lines = lines[1:]

    memory = {}
    for l in lines:
        t, l = l[0], l[1:]

        d = {}
        for k, v in zip(headers, l):
            d[k] = mem_convert(v)
        memory[t] = d

    return memory


def should_run_submake(make_flags):
    """Check if make_flags indicate that we should execute things.

    See https://www.gnu.org/software/make/manual/html_node/Instead-of-Execution.html#Instead-of-Execution  # noqa

    If this is a dry run or question then we shouldn't execute or output
    anything.

    The flags end up as single letter versions in the MAKEFLAGS environment
    variable.

    >>> should_run_submake('')
    True

    The following flags are important;

     -n == --dry-run

    >>> should_run_submake('n')
    False
    >>> should_run_submake('n --blah')
    False
    >>> should_run_submake('--blah n')
    False
    >>> should_run_submake('--blah')
    True
    >>> should_run_submake('--random')
    True

     -q == --question

    >>> should_run_submake('q')
    False
    >>> should_run_submake('q --blah')
    False
    >>> should_run_submake('--blah q')
    False
    >>> should_run_submake('--blah')
    True
    >>> should_run_submake('--random')
    True

      Both --dry-run and --question

    >>> should_run_submake('qn')
    False
    >>> should_run_submake('nq')
    False
    >>> should_run_submake('--quiant')
    True
    """
    r = re.search(r'(?:^|\s)[^-]*(n|q)[^\s]*(\s|$)', make_flags)
    if not r:
        return True
    return not bool(r.groups()[0])


def main(argv):
    fuzzers_dir = os.path.abspath(os.path.dirname(__file__))

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("fuzzer", help="fuzzer to run")
    parser.add_argument(
        "--retries",
        type=int,
        default=0,
        help="Retry a failed fuzzer n times.",
    )
    args = parser.parse_args()

    # Setup the logger with flush=False, it should be safe to use in a signal
    # handler.
    fuzzer_length = max(len(f) for f in os.listdir(fuzzers_dir))
    logger = Logger(args.fuzzer, datetime.utcnow(), fuzzer_length)

    # Need a signal handler on SIGCHLD otherwise get_resource doesn't return
    # anything.
    signal.signal(signal.SIGCHLD, lambda sig, frame: None)

    fuzzer_dir = os.path.join(fuzzers_dir, args.fuzzer)
    assert os.path.exists(fuzzer_dir), fuzzer_dir

    fuzzer_logdir = os.path.join(
        fuzzer_dir, "logs_{}".format(os.environ['XRAY_PART']))
    if not os.path.exists(fuzzer_logdir):
        os.makedirs(fuzzer_logdir)
    assert os.path.exists(fuzzer_logdir)

    # Update the environment for a specific part
    environment = get_environment_variables()
    for key, value in environment.items():
        os.environ[key] = value

    exit_code = -1
    args.retries += 1
    for retry_count in range(0, args.retries):
        logger.log('Running fuzzer attempt: {}', [retry_count])
        exit_code = run_fuzzer(
            args.fuzzer,
            fuzzer_dir,
            fuzzer_logdir,
            logger,
            will_retry=retry_count < (args.retries - 1),
        )
        if exit_code <= 0:
            break
        logger.log('WARNING: Fuzzer failed!')

    return exit_code


def run_fuzzer(fuzzer_name, fuzzer_dir, fuzzer_logdir, logger, will_retry):
    def log(msg, *a, **k):
        logger.log(msg, a, k, flush=True)

    make_cmd = os.environ.get('MAKE', 'make')
    make_flags = os.environ.get('MAKEFLAGS', '')
    # Should run things?
    if not should_run_submake(make_flags):
        return 0

    fuzzer_runok = os.path.join(
        fuzzer_dir, "run.{}.ok".format(os.environ['XRAY_PART']))
    if os.path.exists(fuzzer_runok):
        last_modified = datetime.fromtimestamp(os.stat(fuzzer_runok).st_mtime)

        log(
            "Skipping as run.{}.ok exists (updated @ {})",
            os.environ['XRAY_PART'], last_modified.isoformat())

        return 0

    fuzzer_runok = os.path.join(fuzzer_dir, "run.ok")
    if os.path.exists(fuzzer_runok):
        last_modified = datetime.fromtimestamp(os.stat(fuzzer_runok).st_mtime)

        log(
            "Skipping as run.ok exists (updated @ {})",
            last_modified.isoformat())

        return 0

    time_start = datetime.utcnow()
    log("Starting @ {}", time_start.isoformat())

    running_msg = "Running {} -C {} run (with MAKEFLAGS='{}')".format(
        make_cmd,
        fuzzer_dir,
        make_flags,
    )
    log(running_msg)

    log_suffix = ".{}.log".format(time_start.isoformat()).replace(":", "-")
    fuzzer_stdout = os.path.join(fuzzer_logdir, "stdout" + log_suffix)
    fuzzer_stderr = os.path.join(fuzzer_logdir, "stderr" + log_suffix)

    # Write header to stdout/stderr to make sure they match.
    for fname in [fuzzer_stdout, fuzzer_stderr]:
        with open(fname, "w") as fd:
            fd.write("Build starting @ {}\n".format(time_start.isoformat()))
            fd.write(running_msg)
            fd.write("\n")
            fd.write("-" * 75)
            fd.write("\n")
            fd.flush()
            os.fsync(fd)

    # Open the log files for appending
    stdout_fd = open(fuzzer_stdout, "a")
    stderr_fd = open(fuzzer_stderr, "a")

    # Play nice with make's jobserver.
    # See https://www.gnu.org/software/make/manual/html_node/POSIX-Jobserver.html#POSIX-Jobserver  # noqa
    job_fds = []
    if '--jobserver-fds' in make_flags or '--jobserver-auth' in make_flags:
        job_re = re.search(
            '--jobserver-(?:fds|auth)=([0-9]+),([0-9]+)', make_flags)
        assert job_re, make_flags
        job_rd, job_wr = job_re.groups()

        # Make copies of jobserver FDs in case a retry is needed.

        job_rd = int(job_rd)
        job_wr = int(job_wr)
        assert job_rd > 2, (job_rd, job_wr, make_flags)
        assert job_wr > 2, (job_rd, job_wr, make_flags)
        job_fds.append(job_rd)
        job_fds.append(job_wr)

    p = None
    try:
        p = subprocess.Popen(
            [make_cmd, '-C', fuzzer_dir, 'run'],
            stdin=None,
            stdout=stdout_fd,
            stderr=stderr_fd,
            pass_fds=job_fds,
        )
        while True:
            try:
                retcode = p.wait(timeout=10)
                p = None
            except subprocess.TimeoutExpired:
                retcode = None

            if retcode is not None:
                break
            mem = get_memory()['mem']
            log(
                "Still running (1m:{:0.2f}%, 5m:{:0.2f}%, 15m:{:0.2f}% Mem:{:0.1f}Gi used, {:0.1f}Gi free).\n{}",
                *get_load(),
                mem['used'] / 1e9,
                mem['available'] /
                1e9,  # Using available so the numbers add up.
                PsTree.get(p.pid),
            )
    except (Exception, KeyboardInterrupt, SystemExit):
        retcode = -1
        tb = io.StringIO()
        traceback.print_exc(file=tb)
        log(tb.getvalue())

    # Prevent Ctrl-C so we exit properly...
    old_sigint_handler = signal.getsignal(signal.SIGINT)

    def log_die(sig, frame):
        logger.log("Dieing!")

    signal.signal(signal.SIGINT, log_die)

    # Cleanup child process if they haven't already died.
    try:
        if p is not None:
            try:
                retcode = p.wait(1)
            except subprocess.TimeoutExpired:
                retcode = -1
                p.kill()
                p.wait()
                log("Warning: Killed program which should have been dead!")
    except Exception:
        tb = io.StringIO()
        traceback.print_exc(file=tb)
        log(tb.getvalue())

    # Wait for all children to finish.
    try:
        while True:
            log("Child finished: {}", os.waitpid(-1, 0))
    except ChildProcessError:
        pass

    log("Finishing ({}).", get_usage())

    time_end = datetime.utcnow()

    error_log = "\n".join(last_lines(open(fuzzer_stderr), 10000))
    success_log = "\n".join(last_lines(open(fuzzer_stdout), 100))

    # Find the next X_sponge_log.xml file name...
    for i in range(0, 100):
        tsfilename = os.path.join(fuzzer_logdir, '{}_sponge_log.xml'.format(i))
        if not os.path.exists(tsfilename):
            break

    test_case = junit_xml.TestCase(
        name=fuzzer_name,
        timestamp=time_start.timestamp(),
        elapsed_sec=(time_end - time_start).total_seconds(),
        stdout=success_log,
        stderr=error_log,
    )

    if retcode != 0:
        test_case.add_failure_info(
            'Fuzzer failed with exit code: {}'.format(retcode))

    with open(tsfilename, 'w') as f:
        ts = junit_xml.TestSuite(fuzzer_name, [test_case])
        junit_xml.TestSuite.to_file(f, [ts])

    if retcode != 0:
        test_case.add_failure_info(
            'Fuzzer failed with exit code: {}'.format(retcode), )

        # Log the last 10,000 lines of stderr on a failure
        log(
            """\
--------------------------------------------------------------------------
!Failed! @ {time_end} with exit code: {retcode}
--------------------------------------------------------------------------
- STDOUT (see {stdout_fname} for full log):
--------------------------------------------------------------------------
{stdout_log}
--------------------------------------------------------------------------
- STDERR (see {stderr_fname} for full log):
--------------------------------------------------------------------------
{stderr_log}
--------------------------------------------------------------------------
!Failed! @ {time_end} with exit code: {retcode}
--------------------------------------------------------------------------
""",
            retcode=retcode,
            stdout_fname=fuzzer_stdout,
            stdout_log='\n'.join(last_lines(open(fuzzer_stdout), 1000)),
            stderr_fname=fuzzer_stderr,
            stderr_log='\n'.join(last_lines(open(fuzzer_stderr), 1000)),
            time_end=time_end.isoformat())
    else:
        # Log the last 100 lines of a successful run
        log(
            """\
Succeeded! @ {}
--------------------------------------------------------------------------
{}
--------------------------------------------------------------------------
Succeeded! @ {}
""", time_end.isoformat(), success_log, time_end.isoformat())

    logger.flush()
    signal.signal(signal.SIGINT, old_sigint_handler)
    return retcode


if __name__ == "__main__":
    if "--test" in sys.argv or len(sys.argv) == 1:
        import doctest
        doctest.testmod()
    else:
        sys.exit(main(sys.argv))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Emits top.v's for various BUFHCE routing inputs. """
import os
import random
import re
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray.grid_types import GridLoc
from prjxray.db import Database
from prjxray.lut_maker import LutMaker
from io import StringIO
import csv
import sys

CMT_XY_FUN = util.create_xy_fun(prefix='')
BUFGCTRL_XY_FUN = util.create_xy_fun('BUFGCTRL_')
BUFHCE_XY_FUN = util.create_xy_fun('BUFHCE_')


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def gen_sites(desired_site_type):
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        for site, site_type in gridinfo.sites.items():
            if site_type == desired_site_type:
                yield loc, gridinfo.tile_type, site


def gen_bufhce_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        sites = []

        for site, site_type in gridinfo.sites.items():
            if site_type == 'BUFHCE':
                sites.append(site)

        if sites:
            yield tile_name, sorted(sites)


def get_cmt_loc(cmt_tile_name):
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    return grid.loc_of_tilename(cmt_tile_name)


def read_site_to_cmt():
    """ Yields clock sources and which CMT they route within. """
    with open(os.path.join(os.getenv('FUZDIR'), 'build',
                           'cmt_regions.csv')) as f:
        for l in f:
            site, cmt = l.strip().split(',')
            yield (site, cmt)


def read_pss_clocks():
    with open(os.path.join(os.getenv('FUZDIR'), 'build',
                           'pss_clocks.csv')) as f:
        for l in csv.DictReader(f):
            yield l


class ClockSources(object):
    """ Class for tracking clock sources.

    Some clock sources can be routed to any CMT, for these, cmt='ANY'.
    For clock sources that belong to a CMT, cmt should be set to the CMT of
    the source.

    """

    def __init__(self):
        self.sources = {}
        self.merged_sources = {}
        self.source_to_cmt = {}
        self.used_sources_from_cmt = {}
        self.sources_by_loc = {}
        self.active_cmt_ports = {}

    def add_clock_source(self, source, cmt, loc=None):
        """ Adds a source from a specific CMT.

        cmt='ANY' indicates that this source can be routed to any CMT.
        """
        if cmt not in self.sources:
            self.sources[cmt] = []

        self.sources[cmt].append(source)
        assert source not in self.source_to_cmt or self.source_to_cmt[
            source] == cmt, source
        self.source_to_cmt[source] = cmt

        self.add_bufg_clock_source(source, cmt, loc)

    def add_bufg_clock_source(self, source, cmt, loc):
        if loc not in self.sources_by_loc:
            self.sources_by_loc[loc] = []

        self.sources_by_loc[loc].append((cmt, source))

    def get_random_source(self, cmt):
        """ Get a random source that is routable to the specific CMT.

        get_random_source will return a source that is either cmt='ANY',
        cmt equal to the input CMT, or the adjecent CMT.

        """
        if cmt not in self.merged_sources:
            choices = []
            if 'ANY' in self.sources:
                choices.extend(self.sources['ANY'])

            if cmt in self.sources:
                choices.extend(self.sources[cmt])

            x, y = CMT_XY_FUN(cmt)

            if x % 2 == 0:
                x += 1
            else:
                x -= 1

            paired_cmt = 'X{}Y{}'.format(x, y)

            if paired_cmt in self.sources:
                choices.extend(self.sources[paired_cmt])

            self.merged_sources[cmt] = choices

        if self.merged_sources[cmt]:
            source = random.choice(self.merged_sources[cmt])

            source_cmt = self.source_to_cmt[source]
            if source_cmt not in self.used_sources_from_cmt:
                self.used_sources_from_cmt[source_cmt] = set()

            self.used_sources_from_cmt[source_cmt].add(source)

            if source_cmt != 'ANY' and len(
                    self.used_sources_from_cmt[source_cmt]) > 14:
                self.used_sources_from_cmt[source_cmt].remove(source)
                return None
            else:
                return source

    def get_bufg_source(self, loc, tile_type, site, todos, i_wire, used_only):
        bufg_sources = []

        top = '_TOP_' in tile_type
        bottom = '_BOT_' in tile_type

        assert top ^ bottom, tile_type

        if top:
            for src_loc, cmt_sources in self.sources_by_loc.items():
                if src_loc is None:
                    continue
                if src_loc.grid_y <= loc.grid_y:
                    bufg_sources.extend(cmt_sources)
        elif bottom:
            for src_loc, cmt_sources in self.sources_by_loc.items():
                if src_loc is None:
                    continue
                if src_loc.grid_y > loc.grid_y:
                    bufg_sources.extend(cmt_sources)

        # CLK_HROW_TOP_R_CK_BUFG_CASCO0  -> CLK_BUFG_BUFGCTRL0_I0
        # CLK_HROW_TOP_R_CK_BUFG_CASCO22 -> CLK_BUFG_BUFGCTRL11_I0
        # CLK_HROW_TOP_R_CK_BUFG_CASCO23 -> CLK_BUFG_BUFGCTRL11_I1
        # CLK_HROW_BOT_R_CK_BUFG_CASCO27 -> CLK_BUFG_BUFGCTRL13_I1

        x, y = BUFGCTRL_XY_FUN(site)
        assert x == 0
        y = y % 16

        assert i_wire in [0, 1], i_wire

        casco_wire = '{tile_type}_CK_BUFG_CASCO{casco_idx}'.format(
            tile_type=tile_type.replace('BUFG', 'HROW'),
            casco_idx=(y * 2 + i_wire))

        if casco_wire not in todos:
            return None

        target_wires = []

        need_bufr = False
        for src_wire in todos[casco_wire]:
            if 'BUFRCLK' in src_wire:
                need_bufr = True
                break

        for cmt, wire in bufg_sources:
            if 'BUFR' in wire:
                if need_bufr:
                    target_wires.append((cmt, wire))
            else:
                target_wires.append((cmt, wire))

        random.shuffle(target_wires)
        for cmt, source in target_wires:
            if cmt == 'ANY':
                return source
            else:
                # Make sure to not try to import move than 14 sources from
                # the CMT, there is limited routing.
                if cmt not in self.used_sources_from_cmt:
                    self.used_sources_from_cmt[cmt] = set()

                if source in self.used_sources_from_cmt[cmt]:
                    return source
                elif used_only:
                    continue

                if len(self.used_sources_from_cmt[cmt]) < 14:
                    self.used_sources_from_cmt[cmt].add(source)
                    return source
                else:
                    continue

        return None


def check_allowed(mmcm_pll_dir, cmt):
    """ Check whether the CMT specified is in the allowed direction.

    This function is designed to bias sources to either the left or right
    input lines.

    """
    if mmcm_pll_dir == 'BOTH':
        return True
    elif mmcm_pll_dir == 'ODD':
        x, y = CMT_XY_FUN(cmt)
        return (x & 1) == 1
    elif mmcm_pll_dir == 'EVEN':
        x, y = CMT_XY_FUN(cmt)
        return (x & 1) == 0
    elif mmcm_pll_dir == 'NONE':
        return False
    else:
        assert False, mmcm_pll_dir


def read_todo():
    dsts = {}

    with open(os.path.join('..', 'todo_all.txt')) as f:
        for l in f:
            tile_type, dst, src = l.strip().split('.')

            if dst not in dsts:
                dsts[dst] = set()

            dsts[dst].add(src)

    return dsts


def need_int_connections(todos):
    for srcs in todos.values():
        for src in srcs:
            if re.search('INT_._.', src):
                return True

    return False


def bufhce_in_todo(todos, site):
    if 'BUFHCE' in site:
        # CLK_HROW_CK_MUX_OUT_R9 -> X1Y9
        # CLK_HROW_CK_MUX_OUT_L11 -> X0Y35
        x, y = BUFHCE_XY_FUN(site)
        y = y % 12

        if x == 0:
            lr = 'L'
        elif x == 1:
            lr = 'R'
        else:
            assert False, x

        return 'CLK_HROW_CK_MUX_OUT_{lr}{y}'.format(lr=lr, y=y) in todos
    else:
        return True


def need_gclk_connection(todos, site):
    x, y = BUFGCTRL_XY_FUN(site)
    assert x == 0

    src_wire = 'CLK_HROW_R_CK_GCLK{}'.format(y)
    for srcs in todos.values():
        if src_wire in srcs:
            return True

    return False


def only_gclk_left(todos):
    for srcs in todos.values():
        for src in srcs:
            if 'GCLK' not in src:
                return False

    return True


def main():
    """
    BUFHCE's can be driven from:

        MMCME2_ADV
        PLLE2_ADV
        BUFGCTRL
        Local INT connect
        PS7 (Zynq)
    """

    print('''
// SEED={}
module top();
    '''.format(os.getenv('SEED')))

    is_zynq = os.getenv('XRAY_DATABASE') == 'zynq7'
    clock_sources = ClockSources()

    site_to_cmt = dict(read_site_to_cmt())

    if is_zynq:
        pss_clocks = list(read_pss_clocks())

    # To ensure that all left or right sources are used, sometimes only MMCM/PLL
    # sources are allowed.  The force of ODD/EVEN/BOTH further biases the
    # clock sources to the left or right column inputs.
    mmcm_pll_only = random.randint(0, 1)
    mmcm_pll_dir = random.choice(('ODD', 'EVEN', 'BOTH', 'NONE'))

    todos = read_todo()

    if only_gclk_left(todos):
        mmcm_pll_dir = 'NONE'

    if not mmcm_pll_only:
        if need_int_connections(todos):
            for _ in range(10):
                clock_sources.add_clock_source('one', 'ANY')
                clock_sources.add_clock_source('zero', 'ANY')

    print("""
    wire zero = 0;
    wire one = 1;""")

    for loc, _, site in gen_sites('MMCME2_ADV'):
        mmcm_clocks = [
            'mmcm_clock_{site}_{idx}'.format(site=site, idx=idx)
            for idx in range(13)
        ]

        if check_allowed(mmcm_pll_dir, site_to_cmt[site]):
            for clk in mmcm_clocks:
                clock_sources.add_clock_source(clk, site_to_cmt[site], loc)

        print(
            """
    wire {c0}, {c1}, {c2}, {c3}, {c4}, {c5};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    MMCME2_ADV pll_{site} (
        .CLKOUT0({c0}),
        .CLKOUT0B({c1}),
        .CLKOUT1({c2}),
        .CLKOUT1B({c3}),
        .CLKOUT2({c4}),
        .CLKOUT2B({c5}),
        .CLKOUT3({c6}),
        .CLKOUT3B({c7}),
        .CLKOUT4({c8}),
        .CLKOUT5({c9}),
        .CLKOUT6({c10}),
        .CLKFBOUT({c11}),
        .CLKFBOUTB({c12})
    );
        """.format(
                site=site,
                c0=mmcm_clocks[0],
                c1=mmcm_clocks[1],
                c2=mmcm_clocks[2],
                c3=mmcm_clocks[3],
                c4=mmcm_clocks[4],
                c5=mmcm_clocks[5],
                c6=mmcm_clocks[6],
                c7=mmcm_clocks[7],
                c8=mmcm_clocks[8],
                c9=mmcm_clocks[9],
                c10=mmcm_clocks[10],
                c11=mmcm_clocks[11],
                c12=mmcm_clocks[12],
            ))

    for loc, _, site in gen_sites('PLLE2_ADV'):
        pll_clocks = [
            'pll_clock_{site}_{idx}'.format(site=site, idx=idx)
            for idx in range(6)
        ]

        if check_allowed(mmcm_pll_dir, site_to_cmt[site]):
            for clk in pll_clocks:
                clock_sources.add_clock_source(clk, site_to_cmt[site], loc)

        print(
            """
    wire {c0}, {c1}, {c2}, {c3}, {c4}, {c5};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    PLLE2_ADV pll_{site} (
        .CLKOUT0({c0}),
        .CLKOUT1({c1}),
        .CLKOUT2({c2}),
        .CLKOUT3({c3}),
        .CLKOUT4({c4}),
        .CLKOUT5({c5})
    );
        """.format(
                site=site,
                c0=pll_clocks[0],
                c1=pll_clocks[1],
                c2=pll_clocks[2],
                c3=pll_clocks[3],
                c4=pll_clocks[4],
                c5=pll_clocks[5],
            ))

    for loc, _, site in gen_sites('BUFR'):
        clock_sources.add_bufg_clock_source(
            'O_{site}'.format(site=site), site_to_cmt[site], loc)
        print(
            """
    wire O_{site};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFR bufr_{site} (
        .O(O_{site})
        );""".format(site=site))

    if is_zynq:

        # FCLK clocks. Those are generated by the PS and go directly to one of
        # the CLK_HROW tile.
        clocks = [
            "PSS_FCLKCLK0",
            "PSS_FCLKCLK1",
            "PSS_FCLKCLK2",
            "PSS_FCLKCLK3",
        ]

        loc, _, site = next(gen_sites('PS7'))

        print("")

        # Add clock sources and generate wires
        for wire in clocks:
            clock_info = [d for d in pss_clocks if d["pin"] == wire][0]

            # CMT tile
            cmt_tile = clock_info["tile"]
            cmt_loc = get_cmt_loc(cmt_tile)

            # Add only if the input wire is in the todo list
            dsts = [k for k, v in todos.items() if clock_info["wire"] in v]
            if len(dsts) > 0:

                # Wire source clock region. The PS7 is always left of the
                # CLK_HROW tile, but it does not matter here.
                regions = clock_info["clock_regions"].split()
                regions = sorted([(int(r[1]), int(r[3])) for r in regions])

                # Add the clock source
                cmt = "X{}Y{}".format(regions[0][0], regions[0][1])
                clock_sources.add_clock_source(wire, cmt, cmt_loc)

            print("    wire {};".format(wire))

        print(
            """
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    PS7 ps7_{site} (
        .FCLKCLK({{{fclk3}, {fclk2}, {fclk1}, {fclk0}}})
    );
        """.format(
                site=site,
                fclk0=clocks[0],
                fclk1=clocks[1],
                fclk2=clocks[2],
                fclk3=clocks[3]))

    luts = LutMaker()
    bufhs = StringIO()
    bufgs = StringIO()

    gclks = []
    for _, _, site in sorted(gen_sites("BUFGCTRL"),
                             key=lambda x: BUFGCTRL_XY_FUN(x[2])):
        wire_name = 'gclk_{}'.format(site)
        gclks.append(wire_name)

        include_source = True
        if mmcm_pll_only:
            include_source = False
        elif only_gclk_left(todos):
            include_source = need_gclk_connection(todos, site)

        if include_source:
            clock_sources.add_clock_source(wire_name, 'ANY')

        print("""
    wire {wire_name};
    """.format(wire_name=wire_name))
        print(
            """
    wire I1_{site};
    wire I0_{site};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFGCTRL bufg_{site} (
        .O({wire_name}),
        .S1({s1net}),
        .S0({s0net}),
        .IGNORE1({ignore1net}),
        .IGNORE0({ignore0net}),
        .I1(I1_{site}),
        .I0(I0_{site}),
        .CE1({ce1net}),
        .CE0({ce0net})
        );
        """.format(
                site=site,
                wire_name=wire_name,
                s1net=luts.get_next_output_net(),
                s0net=luts.get_next_output_net(),
                ignore1net=luts.get_next_output_net(),
                ignore0net=luts.get_next_output_net(),
                ce1net=luts.get_next_output_net(),
                ce0net=luts.get_next_output_net(),
            ),
            file=bufgs)

    any_bufhce = False
    for tile_name, sites in gen_bufhce_sites():
        for site in sites:
            if not bufhce_in_todo(todos, site):
                continue

            any_bufhce = True
            print(
                """
    wire I_{site};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFHCE buf_{site} (
        .I(I_{site})
    );
                    """.format(site=site, ),
                file=bufhs)

            if random.random() > .05:
                wire_name = clock_sources.get_random_source(site_to_cmt[site])

                if wire_name is None:
                    continue

                print(
                    """
    assign I_{site} = {wire_name};""".format(
                        site=site,
                        wire_name=wire_name,
                    ),
                    file=bufhs)

    if not any_bufhce:
        for tile_name, sites in gen_bufhce_sites():
            for site in sites:
                print(
                    """
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFHCE #(
        .INIT_OUT({INIT_OUT}),
        .CE_TYPE({CE_TYPE}),
        .IS_CE_INVERTED({IS_CE_INVERTED})
    ) buf_{site} (
        .I({wire_name})
    );
              """.format(
                        INIT_OUT=random.randint(0, 1),
                        CE_TYPE=verilog.quote(
                            random.choice(('SYNC', 'ASYNC'))),
                        IS_CE_INVERTED=random.randint(0, 1),
                        site=site,
                        wire_name=gclks[0],
                    ))
                break
            break

    for l in luts.create_wires_and_luts():
        print(l)

    print(bufhs.getvalue())
    print(bufgs.getvalue())

    used_only = random.random() < .25

    for loc, tile_type, site in sorted(gen_sites("BUFGCTRL"),
                                       key=lambda x: BUFGCTRL_XY_FUN(x[2])):
        if random.randint(0, 1):
            wire_name = clock_sources.get_bufg_source(
                loc, tile_type, site, todos, 1, used_only)
            if wire_name is not None:
                print(
                    """
    assign I1_{site} = {wire_name};""".format(
                        site=site,
                        wire_name=wire_name,
                    ))

        if random.randint(0, 1):
            wire_name = clock_sources.get_bufg_source(
                loc, tile_type, site, todos, 0, used_only)
            if wire_name is not None:
                print(
                    """
    assign I0_{site} = {wire_name};""".format(
                        site=site,
                        wire_name=wire_name,
                    ))

    print("endmodule")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from os import environ, getcwd, chdir, mkdir, environ
import json
from tempfile import TemporaryDirectory
from contextlib import contextmanager
from unittest import TestCase, main

# Setup location of database file to a relative term so it can be generated
# in the current subdirectory, which will be a temporary one, to allow concurent
# testing.
environ['XRAY_DATABASE_ROOT'] = '.'
environ['XRAY_PART'] = 'xc7a200tffg1156-1'

from prjxray.util import get_roi, get_db_root
from prjxray.overlay import Overlay
from prjxray.grid_types import GridLoc


@contextmanager
def setup_database(contents):
    with TemporaryDirectory() as d:
        olddir = getcwd()
        chdir(d)
        mkdir('xc7a200t')
        mkdir('mapping')
        environ['XRAY_DATABASE_ROOT'] = d
        e = None
        with open('xc7a200t/tilegrid.json', 'w') as fd:
            json.dump(contents, fd)
        # Create some dummy data
        with open('mapping/devices.yaml', 'w') as fd:
            json.dump({'xc7a200t': {'fabric': "xc7a200t"}}, fd)
        with open('mapping/parts.yaml', 'w') as fd:
            json.dump({'xc7a200tffg1156-1': {"device": "xc7a200t"}}, fd)

        try:
            yield
        except Exception as ereal:
            e = ereal
        chdir(olddir)
        if e is not None:
            raise e


class TestUtil(TestCase):
    def test_get_roi_gen_sites(self):
        makedb = lambda sites: {
                "ATILE": {
                    "bits": {
                        "CLB_IO_CLK": {
                            "baseaddr": "0x00400F00",
                            "frames": 28,
                            "height": 2,
                            "offset": 0,
                            "words": 2
                        }
                    },
                    "grid_x": 10,
                    "grid_y": 10,
                    "segment": "ASEGMENT",
                    "segment_type": "bram0_l",
                    "sites": sites,
                    "prohibited_sites": [],
                    "type": "BRAM_INT_INTERFACE_L"
                }
            }
        with setup_database(makedb({})):
            self.assertListEqual(list(get_roi().gen_sites()), [])
        with setup_database(makedb({'FOO': 'BAR'})):
            self.assertListEqual(
                list(get_roi().gen_sites()), [('ATILE', 'FOO', 'BAR')])

    def test_in_roi_overlay(self):
        region_dict = {}
        region_dict['pr1'] = (10, 58, 0, 51)
        region_dict['pr2'] = (10, 58, 52, 103)
        overlay = Overlay(region_dict)
        self.assertFalse(overlay.tile_in_roi(GridLoc(18, 50)))
        self.assertFalse(overlay.tile_in_roi(GridLoc(18, 84)))
        self.assertTrue(overlay.tile_in_roi(GridLoc(8, 50)))
        self.assertTrue(overlay.tile_in_roi(GridLoc(18, 112)))
        self.assertTrue(overlay.tile_in_roi(GridLoc(80, 40)))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker
import os
import os.path


def bitfilter(frame, word):
    if frame < 26:
        return False

    return True


IOCLK_MAP = {
    'HCLK_IOI_I2IOCLK_TOP0': 'HCLK_CMT_CCIO0',
    'HCLK_IOI_I2IOCLK_TOP1': 'HCLK_CMT_CCIO1',
    'HCLK_IOI_I2IOCLK_BOT0': 'HCLK_CMT_CCIO2',
    'HCLK_IOI_I2IOCLK_BOT1': 'HCLK_CMT_CCIO3',
}

IOCLK_SRCS = set(IOCLK_MAP.values())


def main():
    segmk = Segmaker("design.bits")

    tiledata = {}
    pipdata = {}
    ignpip = set()
    tile_ports = {}

    with open(os.path.join(os.getenv('FUZDIR'), '..', 'piplist', 'build',
                           'hclk_cmt', 'hclk_cmt.txt')) as f:
        for l in f:
            tile_type, dst, src = l.strip().split('.')
            if tile_type not in pipdata:
                pipdata[tile_type] = []
                tile_ports[tile_type] = set()

            pipdata[tile_type].append((src, dst))
            tile_ports[tile_type].add(src)
            tile_ports[tile_type].add(dst)

    with open(os.path.join(os.getenv('FUZDIR'), '..', 'piplist', 'build',
                           'hclk_cmt', 'hclk_cmt_l.txt')) as f:
        for l in f:
            tile_type, dst, src = l.strip().split('.')
            if tile_type not in pipdata:
                pipdata[tile_type] = []
                tile_ports[tile_type] = set()

            pipdata[tile_type].append((src, dst))
            tile_ports[tile_type].add(src)
            tile_ports[tile_type].add(dst)

    tile_to_cmt = {}
    cmt_to_hclk_cmt = {}
    with open(os.path.join(os.getenv('FUZDIR'), 'build',
                           'cmt_regions.csv')) as f:
        for l in f:
            site, cmt, tile = l.strip().split(',')

            tile_to_cmt[tile] = cmt

            if tile.startswith('HCLK_CMT'):
                cmt_to_hclk_cmt[cmt] = tile

    active_ioclks = set()

    print("Loading tags from design.txt.")
    with open("design.txt", "r") as f:
        for line in f:
            tile, pip, src, dst, pnum, pdir = line.split()

            pip_prefix, _ = pip.split(".")
            tile_from_pip, tile_type = pip_prefix.split('/')
            assert tile == tile_from_pip
            _, src = src.split("/")
            _, dst = dst.split("/")
            pnum = int(pnum)
            pdir = int(pdir)

            if src in IOCLK_MAP:
                active_ioclks.add(
                    (cmt_to_hclk_cmt[tile_to_cmt[tile]], IOCLK_MAP[src]))

            if not tile.startswith('HCLK_CMT'):
                continue

            if tile not in tiledata:
                tiledata[tile] = {
                    "type": tile_type,
                    "pips": set(),
                    "srcs": set(),
                    "dsts": set()
                }

            tiledata[tile]["pips"].add((src, dst))
            tiledata[tile]["srcs"].add(src)
            tiledata[tile]["dsts"].add(dst)

            if pdir == 0:
                tiledata[tile]["srcs"].add(dst)
                tiledata[tile]["dsts"].add(src)

            if pnum == 1 or pdir == 0:
                ignpip.add((src, dst))

    for tile, pips_srcs_dsts in tiledata.items():
        tile_type = pips_srcs_dsts["type"]
        pips = pips_srcs_dsts["pips"]

        for src, dst in pipdata[tile_type]:
            if (src, dst) in ignpip:
                pass
            elif (src, dst) in pips:
                segmk.add_tile_tag(tile, "%s.%s" % (dst, src), 1)
            elif dst not in tiledata[tile]["dsts"]:
                segmk.add_tile_tag(tile, "%s.%s" % (dst, src), 0)

        for port in tile_ports[tile_type]:

            # These ones do not have any outgoing connections from the tile.
            if "FREQ_REF" in port:
                continue

            # There seems to be no special bits related to use of
            # HCLK_CMT_MUX_CLKINT_n wires.
            if "HCLK_CMT_MUX_CLKINT" in port:
                continue

            # It seems that CCIOn_USED is not enabled when a net goes through
            # FREQ_REFn. Do not emit this tag if this happens.
            if "CCIO" in port:
                n = int(port[-1])
                dst = "HCLK_CMT_MUX_OUT_FREQ_REF{}".format(n)
                if dst in tiledata[tile]["dsts"]:
                    continue

            if port in tiledata[tile]["dsts"] or port in tiledata[tile]["srcs"]:
                segmk.add_tile_tag(tile, "{}_USED".format(port), 1)
            else:
                segmk.add_tile_tag(tile, "{}_USED".format(port), 0)

        for ioclk in IOCLK_SRCS:
            if ioclk in tiledata[tile]["srcs"] or (tile,
                                                   ioclk) in active_ioclks:
                segmk.add_tile_tag(tile, "{}_ACTIVE".format(ioclk), 1)
            else:
                segmk.add_tile_tag(tile, "{}_ACTIVE".format(ioclk), 0)

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json
import os
from enum import Enum

from prjxray.segmaker import Segmaker
#from icecream import ic

INT = "INT"
BIN = "BIN"


def bitfilter_gtx_common_mid(frame, bit):
    # Filter out non interesting bits.
    word = int(bit / 32)

    if word < 44 or word > 56:
        return False

    if frame not in [0, 1]:
        return False

    return True


def bitfilter_gtx_common(frame, bit):
    # Filter out non interesting bits.
    word = int(bit / 32)

    if word < 44 or word > 56:
        return False

    # let ENABLE_DRP come through
    if (frame == 24 or frame or frame == 25) and bit == 1613:
        return True
    
    if frame < 30 or frame > 31:
        return False

    return True


def main():
    segmk = Segmaker("design.bits")

    fuz_dir = os.getenv("FUZDIR", None)
    assert fuz_dir
    with open(os.path.join(fuz_dir, "attrs.json"), "r") as attr_file:
        attrs = json.load(attr_file)

    print("Loading tags")
    with open("params.json") as f:
        params_dict = json.load(f)
        tile_type = params_dict["tile_type"]
        params_list = params_dict["params"]

    sites_in_tile = dict()

    for params in params_list:
        site = params["site"]
        tile = params["tile"]
        
        if "GTXE2_COMMON" not in site:
            continue

        sites_in_tile[tile] = site

        in_use = params["IN_USE"]

        segmk.add_site_tag(site, "IN_USE", in_use)

        if in_use:
            for param, param_info in attrs.items():
                value = params[param]
                param_type = param_info["type"]
                param_digits = param_info["digits"]
                param_values = param_info["values"]

                if param_type == INT:
                    param_encodings = param_info["encoding"]
                    param_encoding = param_encodings[param_values.index(value)]
                    bitstr = [
                        int(x) for x in "{value:0{digits}b}".format(
                            value=param_encoding, digits=param_digits)[::-1]
                    ]

                    for i in range(param_digits):
                        segmk.add_site_tag(
                            site, '%s[%u]' % (param, i), bitstr[i])
                else:
                    assert param_type == BIN
                    bitstr = [
                        int(x) for x in "{value:0{digits}b}".format(
                            value=value, digits=param_digits)[::-1]
                    ]

                    for i in range(param_digits):
                        segmk.add_site_tag(
                            site, "%s[%u]" % (param, i), bitstr[i])

            for param in ["QPLLLOCKDETCLK", "DRPCLK"]:
                segmk.add_site_tag(site, "INV_" + param, params[param])

            for param in ["GTREFCLK0_USED", "GTREFCLK1_USED",
                          "BOTH_GTREFCLK_USED"]:
                segmk.add_site_tag(site, param, params[param])

            segmk.add_site_tag(site, "ENABLE_DRP", params["ENABLE_DRP"])

    for params in params_list:
        site = params["site"]

        if "IBUFDS_GTE2" not in site:
            continue

        in_use = params["IN_USE"]
        segmk.add_site_tag(site, "IN_USE", in_use)

        if in_use:
            tile = params["tile"]

            for param in ["CLKRCV_TRST", "CLKCM_CFG"]:
                value = params[param]
                segmk.add_site_tag(site, param, "TRUE" in value)

            bitstr = [
                int(x) for x in "{value:0{digits}b}".format(
                    value=params["CLKSWING_CFG"], digits=2)[::-1]
            ]

            gtx_common_site = sites_in_tile[tile]
            for i in range(2):
                segmk.add_site_tag(
                    gtx_common_site, "IBUFDS_GTE2.CLKSWING_CFG[%u]" % (i),
                    bitstr[i])

    if tile_type.startswith("GTX_COMMON_MID"):
        bitfilter = bitfilter_gtx_common_mid
    elif tile_type == "GTX_COMMON":
        bitfilter = bitfilter_gtx_common
    else:
        assert False, tile_type

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json

from prjxray.segmaker import Segmaker
from prjxray import verilog
from prjxray import segmaker


def bitfilter(frame, word):
    if frame < 26:
        return False
    return True


def run():

    segmk = Segmaker("design.bits")

    print("Loading tags")
    f = open('params.jl', 'r')
    design = json.load(f)
    for p in design:
        ps = p["params"]
        if p["site_type"] in "ICAP":
            param = verilog.unquote(ps["ICAP_WIDTH"])
            segmaker.add_site_group_zero(
                segmk, p["site"], "ICAP_WIDTH_", ["X32", "X8", "X16"], "X32",
                param)
        elif p["site_type"] in "BSCAN":
            param = str(ps["JTAG_CHAIN"])
            segmaker.add_site_group_zero(
                segmk, p["site"], "JTAG_CHAIN_", ["1", "2", "3", "4"], param,
                param)
        elif p["site_type"] in "CAPTURE":
            param = verilog.unquote(ps["ONESHOT"])
            segmk.add_site_tag(
                p["site"], "ONESHOT", True if param in "TRUE" else False)
        elif p["site_type"] in "STARTUP":
            param = verilog.unquote(ps["PROG_USR"])
            segmk.add_site_tag(
                p["site"], "PROG_USR", True if param in "TRUE" else False)
        elif p["site_type"] in "FRAME_ECC":
            param = verilog.unquote(ps["FARSRC"])
            segmaker.add_site_group_zero(
                segmk, p["site"], "FARSRC_", ["FAR", "EFAR"], param, param)
        elif p["site_type"] in ["USR_ACCESS", "DCIRESET"]:
            feature = "ENABLED"
            segmk.add_site_tag(
                p["site"], feature, True if ps["ENABLED"] else False)

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
This tool allows to view segbits using a 2D frame vs. bit plot. It can read
either mask.db files or segbits.db files. Multiple files can be read at once
allowing to identify where a bit comes from.

When a single file is read, a bit is marked as "O". When multiple files are
read, a bit is denoted with a letter "A", "B" and so on depending on index
of file that it belongs to.

Duplicate bits (present in more than one files) are marked with "#"
"""

import sys
import argparse
import re

from prjxray.util import OpenSafeFile

# =============================================================================


def load_just_bits(file_name):
    """
    Read bits from a .db or .rdb file. Ignores tags and bit values.
    """

    with OpenSafeFile(file_name, "r") as fp:
        lines = fp.readlines()

    bits = set()
    for line in lines:
        for word in line.split(" "):
            match = re.match("^(!?)([0-9]+)_([0-9]+)$", word)
            if match is not None:
                frm = int(match.group(2))
                bit = int(match.group(3))

                bits.add((
                    frm,
                    bit,
                ))

    return bits


# =============================================================================


def main():

    # Colors for TTY
    if sys.stdout.isatty():

        bit_colors = [
            "\033[39m",
            "\033[91m",
            "\033[92m",
            "\033[93m",
            "\033[94m",
            "\033[95m",
            "\033[96m",
            "\033[31m",
            "\033[32m",
            "\033[33m",
            "\033[34m",
            "\033[35m",
            "\033[36m",
        ]

        colors = {
            "NONE": "\033[0m",
            "DUPLICATE": "\033[101;97m",
        }

    # Colors for pipe
    else:

        bit_colors = [""]

        colors = {
            "NONE": "",
            "DUPLICATE": "",
        }

    # .........................................................................

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("files", nargs="*", type=str, help="Input files")

    args = parser.parse_args()

    # Load bits
    all_bits = []
    for i, f in enumerate(args.files):
        bits = load_just_bits(f)
        all_bits.append(bits)

        cstr = bit_colors[i % len(bit_colors)]
        bstr = "O" if len(args.files) == 1 else chr(65 + i)
        print(cstr + bstr + colors["NONE"] + ": %s #%d" % (f, len(bits)))

    print("")

    max_frames = max([bit[0] for bits in all_bits for bit in bits]) + 1
    max_bits = max([bit[1] for bits in all_bits for bit in bits]) + 1

    # Header
    for r in range(3):
        line = " " * 3
        for c in range(max_bits):
            bstr = "%03d" % c
            line += bstr[r]
        print(line)

    print("")

    # Display bits
    for r in range(max_frames):
        line = "%2d " % r
        for c in range(max_bits):
            got_bit = False
            bit_str = colors["NONE"] + "-"
            for i, bits in enumerate(all_bits):
                cstr = bit_colors[i % len(bit_colors)]
                bstr = "O" if len(args.files) == 1 else chr(65 + i)
                if (r, c) in bits:
                    if not got_bit:
                        bit_str = cstr + bstr
                    else:
                        bit_str = colors["DUPLICATE"] + "#" + colors["NONE"]
                    got_bit = True

            line += bit_str

        line += colors["NONE"]
        print(line)


# =============================================================================

if __name__ == "__main__":
    main()
